"""Tests for ConnectedDriveAccount."""
import re
import unittest

import requests_mock

from bimmer_connected.vehicle import CarBrand, DriveTrainType, VehicleViewDirection

from . import (
    VIN_F11,
    VIN_F31,
    VIN_F35,
    VIN_F44,
    VIN_F45,
    VIN_F48,
    VIN_G01,
    VIN_G05,
    VIN_G08,
    VIN_G21,
    VIN_G30,
    VIN_I01_NOREX,
    VIN_I01_REX,
)
from .test_account import get_mocked_account

ATTRIBUTE_MAPPING = {
    "remainingFuel": "remaining_fuel",
    "position": "gps_position",
    "cbsData": "condition_based_services",
    "checkControlMessages": "check_control_messages",
    "doorLockState": "door_lock_state",
    "updateReason": "last_update_reason",
    "chargingLevelHv": "charging_level_hv",
    "chargingStatus": "charging_status",
    "maxRangeElectric": "max_range_electric",
    "remainingRangeElectric": "remaining_range_electric",
    "parkingLight": "parking_lights",
    "remainingRangeFuel": "remaining_range_fuel",
    "updateTime": "timestamp",
    "chargingTimeRemaining": "charging_time_remaining",
}


class TestVehicle(unittest.TestCase):
    """Tests for ConnectedDriveVehicle."""

    def test_drive_train(self):
        """Tests around drive_train attribute."""
        vehicle = get_mocked_account().get_vehicle(VIN_G21)
        self.assertEqual(DriveTrainType.PLUGIN_HYBRID, vehicle.drive_train)

    def test_parsing_attributes(self):
        """Test parsing different attributes of the vehicle."""
        account = get_mocked_account()

        for vehicle in account.vehicles:
            print(vehicle.name)
            self.assertIsNotNone(vehicle.drive_train)
            self.assertIsNotNone(vehicle.name)
            self.assertIsInstance(vehicle.brand, CarBrand)
            self.assertIsNotNone(vehicle.has_internal_combustion_engine)
            self.assertIsNotNone(vehicle.has_hv_battery)
            self.assertIsNotNone(vehicle.drive_train_attributes)
            self.assertIsNotNone(vehicle.has_weekly_planner_service)

    def test_drive_train_attributes(self):
        """Test parsing different attributes of the vehicle."""
        account = get_mocked_account()

        vehicle_drivetrains = {
            VIN_F11: (True, False, False),
            VIN_F31: (True, False, False),
            VIN_F35: (True, False, False),
            VIN_F44: (True, False, False),
            VIN_F45: (True, True, False),
            VIN_F48: (True, False, False),
            VIN_G01: (True, True, False),
            VIN_G05: (True, True, False),
            VIN_G08: (False, True, False),
            VIN_G21: (True, True, False),
            VIN_G30: (True, True, False),
            VIN_I01_NOREX: (False, True, False),
            VIN_I01_REX: (False, True, True),
        }

        for vehicle in account.vehicles:
            self.assertEqual(vehicle_drivetrains[vehicle.vin][0], vehicle.has_internal_combustion_engine)
            self.assertEqual(vehicle_drivetrains[vehicle.vin][1], vehicle.has_hv_battery)
            self.assertEqual(vehicle_drivetrains[vehicle.vin][2], vehicle.has_range_extender)

    def test_parsing_of_lsc_type(self):
        """Test parsing the lsc type field."""
        account = get_mocked_account()

        for vehicle in account.vehicles:
            self.assertIsNotNone(vehicle.lsc_type)

    def test_car_brand(self):
        """Test CarBrand enum"""
        self.assertEqual(CarBrand("BMW"), CarBrand("bmw"))

        with self.assertRaises(ValueError):
            CarBrand("Audi")

    def test_vehicle_attribute_getter(self):
        """Test generic getter."""
        vehicle = get_mocked_account().get_vehicle(VIN_G21)

        self.assertEqual(True, vehicle.has_internal_combustion_engine)

        self.assertEqual("CONNECTED", vehicle.connection_status)

    def test_set_observer_position(self):
        """Test setting observer position"""
        vehicle = get_mocked_account().get_vehicle(VIN_G21)

        vehicle.set_observer_position(12.1, 10.1)

        with self.assertRaises(ValueError):
            vehicle.set_observer_position(12.1, None)

    def test_get_is_tracking_enabled(self):
        """Test setting observer position"""
        vehicle = get_mocked_account().get_vehicle(VIN_F11)
        self.assertEqual(False, vehicle.is_vehicle_tracking_enabled)

        vehicle = get_mocked_account().get_vehicle(VIN_F31)
        self.assertEqual(True, vehicle.is_vehicle_tracking_enabled)

    def test_available_attributes(self):
        """Check that available_attributes returns exactly the arguments we have in our test data."""
        account = get_mocked_account()

        vehicle = account.get_vehicle(VIN_F31)
        self.assertListEqual(["gps_position", "vin"], vehicle.available_attributes)

        vehicle = account.get_vehicle(VIN_G08)
        self.assertListEqual(
            [
                "gps_position",
                "vin",
                "remaining_range_total",
                "mileage",
                "charging_time_remaining",
                "charging_start_time",
                "charging_end_time",
                "charging_time_label",
                "charging_status",
                "charging_level_hv",
                "connection_status",
                "remaining_range_electric",
                "last_charging_end_result",
                "condition_based_services",
                "check_control_messages",
                "door_lock_state",
                "timestamp",
                "last_update_reason",
                "lids",
                "windows",
            ],
            vehicle.available_attributes,
        )

        vehicle = account.get_vehicle(VIN_G30)
        self.assertListEqual(
            [
                "gps_position",
                "vin",
                "remaining_range_total",
                "mileage",
                "charging_time_remaining",
                "charging_start_time",
                "charging_end_time",
                "charging_time_label",
                "charging_status",
                "charging_level_hv",
                "connection_status",
                "remaining_range_electric",
                "last_charging_end_result",
                "remaining_fuel",
                "remaining_range_fuel",
                "fuel_percent",
                "condition_based_services",
                "check_control_messages",
                "door_lock_state",
                "timestamp",
                "last_update_reason",
                "lids",
                "windows",
            ],
            vehicle.available_attributes,
        )

        #     available_attributes = vehicle.available_attributes
        #     print()
        # for vin, dirname in TEST_VEHICLE_DATA.items():
        #     vehicle = account.get_vehicle(vin)
        #     print(vehicle.name)
        #     status_data = load_response_json('{}/status.json'.format(dirname))
        #     existing_attributes = status_data['vehicleStatus'].keys()
        #     existing_attributes = sorted([ATTRIBUTE_MAPPING.get(a, a) for a in existing_attributes
        #                                   if a not in MISSING_ATTRIBUTES])
        #     expected_attributes = sorted([a for a in vehicle.available_attributes if a not in ADDITIONAL_ATTRIBUTES])
        #     self.assertListEqual(existing_attributes, expected_attributes)

    def test_available_state_services(self):
        """Check that available_attributes returns exactly the arguments we have in our test data."""
        vehicle = get_mocked_account().get_vehicle(VIN_F31)

        self.assertListEqual(["status"], vehicle.available_state_services)

    def test_vehicle_image(self):
        """Test vehicle image request."""
        vehicle = get_mocked_account().get_vehicle(VIN_G05)

        with requests_mock.Mocker() as mock:
            mock.get(
                re.compile(r"/eadrax-ics/v3/presentation/vehicles/\w*/images\?carView=\w*"),
                headers={"accept": "image/png"},
                text="png_image",
            )

            self.assertEqual(b"png_image", vehicle.get_vehicle_image(VehicleViewDirection.FRONT))



"""Models state and remote services of one vehicle."""
from enum import Enum
import logging
from typing import TYPE_CHECKING, List

from bimmer_connected.charging_profile import ChargingProfile
from bimmer_connected.vehicle_status import VehicleStatus
from bimmer_connected.remote_services import RemoteServices
from bimmer_connected.const import SERVICE_PROPERTIES, SERVICE_STATUS, VEHICLE_IMAGE_URL
from bimmer_connected.utils import SerializableBaseClass, get_class_property_names, serialize_for_json

if TYPE_CHECKING:
    from bimmer_connected.account import ConnectedDriveAccount

_LOGGER = logging.getLogger(__name__)


class DriveTrainType(str, Enum):
    """Different types of drive trains."""
    COMBUSTION = 'COMBUSTION'
    PLUGIN_HYBRID = 'PLUGIN_HYBRID'  # PHEV
    ELECTRIC = 'ELECTRIC'
    HYBRID = 'HYBRID'  # mild hybrids


class CarBrand(str, Enum):
    """Car brands supported by the My BMW API."""
    @classmethod
    def _missing_(cls, value):
        for member in cls:
            if member.value == value.lower():
                return member
        raise ValueError("'{}' is not a valid {}".format(value, cls.__name__))

    BMW = "bmw"
    MINI = "mini"


#: Set of drive trains that have a combustion engine
COMBUSTION_ENGINE_DRIVE_TRAINS = {DriveTrainType.COMBUSTION, DriveTrainType.PLUGIN_HYBRID}

#: set of drive trains that have a high voltage battery
HV_BATTERY_DRIVE_TRAINS = {DriveTrainType.PLUGIN_HYBRID, DriveTrainType.ELECTRIC}


class VehicleViewDirection(str, Enum):
    """Viewing angles for the vehicle.

    This is used to get a rendered image of the vehicle.
    """
    FRONTSIDE = 'VehicleStatus'
    FRONT = 'VehicleInfo'
    # REARSIDE = 'REARSIDE'
    # REAR = 'REAR'
    SIDE = 'ChargingHistory'
    # DASHBOARD = 'DASHBOARD'
    # DRIVERDOOR = 'DRIVERDOOR'
    # REARBIRDSEYE = 'REARBIRDSEYE'


class LscType(str, Enum):
    """Known Values for lsc_type field.

    Not really sure, what this value really contains.
    """
    NOT_CAPABLE = 'NOT_CAPABLE'
    NOT_SUPPORTED = 'NOT_SUPPORTED'
    ACTIVATED = 'ACTIVATED'


class ConnectedDriveVehicle(SerializableBaseClass):
    """Models state and remote services of one vehicle.

    :param account: ConnectedDrive account this vehicle belongs to
    :param attributes: attributes of the vehicle as provided by the server
    """

    def __init__(self, account: "ConnectedDriveAccount", vehicle_dict: dict) -> None:
        self._account = account
        self.attributes = None
        self.status = VehicleStatus(self._account)
        self.remote_services = RemoteServices(self._account, self)
        self.observer_latitude = None  # type: float
        self.observer_longitude = None  # type: float

        self.update_state(vehicle_dict)

    def update_state(self, vehicle_dict) -> None:
        """Update the state of a vehicle."""
        if SERVICE_STATUS in vehicle_dict and SERVICE_PROPERTIES in vehicle_dict:
            self.attributes = {k: v for k, v in vehicle_dict.items() if k not in [SERVICE_STATUS, SERVICE_PROPERTIES]}
            self.status.update_state(
                {
                    k: v
                    for k, v
                    in vehicle_dict.items()
                    if k in [SERVICE_STATUS, SERVICE_PROPERTIES]
                }
            )
        else:
            _LOGGER.warning("Incomplete vehicle status data: %s", vehicle_dict)

    @property
    def charging_profile(self) -> ChargingProfile:
        """Return the charging profile if available."""
        return ChargingProfile(self.status) if self.has_weekly_planner_service else None

    @property
    def drive_train(self) -> DriveTrainType:
        """Get the type of drive train of the vehicle."""
        return DriveTrainType(self.attributes['driveTrain'])

    @property
    def name(self) -> str:
        """Get the name of the vehicle."""
        return self.attributes['model']

    @property
    def brand(self) -> CarBrand:
        """Get the car brand."""
        return CarBrand(self.attributes["brand"])

    @property
    def has_hv_battery(self) -> bool:
        """Return True if vehicle is equipped with a high voltage battery.

        In this case we can get the state of the battery in the state attributes.
        """
        return self.drive_train in HV_BATTERY_DRIVE_TRAINS

    @property
    def has_range_extender(self) -> bool:
        """Return True if vehicle is equipped with a range extender.

        In this case we can get the state of the gas tank."""
        return self.drive_train == DriveTrainType.ELECTRIC and self.status.fuel_indicator_count == 3

    @property
    def has_internal_combustion_engine(self) -> bool:
        """Return True if vehicle is equipped with an internal combustion engine.

        In this case we can get the state of the gas tank."""
        return self.drive_train in COMBUSTION_ENGINE_DRIVE_TRAINS

    @property
    def has_weekly_planner_service(self) -> bool:
        """Return True if charging control (weekly planner) is available."""
        return self.attributes["capabilities"]["isChargingPlanSupported"]

    @property
    def is_vehicle_tracking_enabled(self) -> bool:
        """Return True if vehicle finder is enabled in vehicle."""
        return self.attributes["capabilities"]["vehicleFinder"]["isEnabled"]

    @property
    def drive_train_attributes(self) -> List[str]:
        """Get list of attributes available for the drive train of the vehicle.

        The list of available attributes depends if on the type of drive train.
        Some attributes only exist for electric/hybrid vehicles, others only if you
        have a combustion engine. Depending on the state of the vehicle, some of
        the attributes might still be None.
        """
        result = ['remaining_range_total', 'mileage']
        if self.has_hv_battery:
            result += ['charging_time_remaining', 'charging_start_time', 'charging_end_time', 'charging_time_label',
                       'charging_status', 'charging_level_hv', 'connection_status', 'remaining_range_electric',
                       'last_charging_end_result']
        if self.has_internal_combustion_engine or self.has_range_extender:
            result += ['remaining_fuel', 'remaining_range_fuel', 'fuel_percent']
        return result

    @property
    def lsc_type(self) -> LscType:
        """Get the lscType of the vehicle.

        Not really sure what that value really means. If it is NOT_CAPABLE, that probably means that the
        vehicle state will not contain much data.
        """
        return LscType(self.attributes["capabilities"]["lastStateCall"].get('lscState'))

    @property
    def available_attributes(self) -> List[str]:
        """Get the list of non-drivetrain attributes available for this vehicle."""
        # attributes available in all vehicles
        result = ['gps_position', 'vin']
        if self.lsc_type == LscType.ACTIVATED:
            # generic attributes if lsc_type =! NOT_SUPPORTED
            result += self.drive_train_attributes
            result += ['condition_based_services', 'check_control_messages', 'door_lock_state', 'timestamp',
                       'last_update_reason']
            # required for existing Home Assistant binary sensors
            result += ['lids', 'windows']
        return result

    @property
    def available_state_services(self) -> List[str]:
        """Get the list of all available state services for this vehicle."""
        result = [SERVICE_STATUS]

        return result

    def get_vehicle_image(self, direction: VehicleViewDirection) -> bytes:
        """Get a rendered image of the vehicle.

        :returns bytes containing the image in PNG format.
        """
        url = VEHICLE_IMAGE_URL.format(
            vin=self.vin,
            server=self._account.server_url,
            view=direction.value,
        )
        header = self._account.request_header()
        # the accept field of the header needs to be updated as we want a png not the usual JSON
        header['accept'] = 'image/png'
        response = self._account.send_request(url, headers=header)
        return response.content

    def __getattr__(self, item):
        """In the first version: just get the attributes from the dict.

        In a later version we might parse the attributes to provide a more advanced API.
        :param item: item to get, as defined in VEHICLE attributes
        """
        if item in get_class_property_names(self):
            return getattr(self, item)
        if item in get_class_property_names(self.status):
            return getattr(self.status, item)
        return self.attributes.get(item)

    def __str__(self) -> str:
        """Use the name as identifier for the vehicle."""
        return '{}: {}'.format(self.__class__, self.name)

    def as_dict(self) -> dict:
        """Return all attributes and parameters, without `self.remote_services`."""
        return serialize_for_json(self, ["remote_services"])

    def set_observer_position(self, latitude: float, longitude: float) -> None:
        """Set the position of the observer, who requests the vehicle state.

        Some vehicle require you to send your position to the server before you get the vehicle state.
        Your position must be within some range (2km?) of the vehicle to get you a proper answer.
        """
        if latitude is None or longitude is None:
            raise ValueError('Either latitude and longitude are both not None or both are None.')
        self.observer_latitude = latitude
        self.observer_longitude = longitude



"""Models the state of a vehicle."""

import datetime
import logging
from enum import Enum
from typing import Dict, List, Optional, Tuple, TYPE_CHECKING

from bimmer_connected.coord_convert import gcj2wgs

from bimmer_connected.country_selector import Regions
from bimmer_connected.utils import SerializableBaseClass, parse_datetime

if TYPE_CHECKING:
    from bimmer_connected.account import ConnectedDriveAccount

_LOGGER = logging.getLogger(__name__)


class LidState(str, Enum):
    """Possible states of the hatch, trunk, doors, windows, sun roof."""
    CLOSED = 'CLOSED'
    OPEN = 'OPEN'
    OPEN_TILT = 'OPEN_TILT'
    INTERMEDIATE = 'INTERMEDIATE'
    INVALID = 'INVALID'


class LockState(str, Enum):
    """Possible states of the door locks."""
    LOCKED = 'LOCKED'
    SECURED = 'SECURED'
    SELECTIVE_LOCKED = 'SELECTIVE_LOCKED'
    UNLOCKED = 'UNLOCKED'
    UNKNOWN = 'UNKNOWN'


class ConditionBasedServiceStatus(str, Enum):
    """Status of the condition based services."""
    OK = 'OK'
    OVERDUE = 'OVERDUE'
    PENDING = 'PENDING'


class ChargingState(str, Enum):
    """Charging state of electric vehicle."""
    DEFAULT = 'DEFAULT'
    CHARGING = 'CHARGING'
    ERROR = 'ERROR'
    COMPLETE = 'COMPLETE'
    FULLY_CHARGED = 'FULLY_CHARGED'
    FINISHED_FULLY_CHARGED = 'FINISHED_FULLY_CHARGED'
    FINISHED_NOT_FULL = 'FINISHED_NOT_FULL'
    INVALID = 'INVALID'
    NOT_CHARGING = 'NOT_CHARGING'
    PLUGGED_IN = 'PLUGGED_IN'
    WAITING_FOR_CHARGING = 'WAITING_FOR_CHARGING'


class CheckControlMessage(SerializableBaseClass):
    """Check control message sent from the server.

    This class provides a nicer API than parsing the JSON format directly.
    """

    def __init__(self, ccm_dict: dict):
        self._ccm_dict = ccm_dict

    @property
    def description_long(self) -> str:
        """Long description of the check control message."""
        return self._ccm_dict.get("longDescription")

    @property
    def description_short(self) -> str:
        """Short description of the check control message."""
        return self._ccm_dict.get("title")

    @property
    def ccm_id(self) -> int:
        """id of the check control message."""
        return self._ccm_dict.get("id")

    @property
    def state(self) -> int:
        """state of the check control message."""
        return self._ccm_dict.get("state")


class FuelIndicator(SerializableBaseClass):
    """Parsed fuel indicators.

    This class provides a nicer API than parsing the JSON format directly.
    """

    # pylint: disable=too-few-public-methods, too-many-instance-attributes

    def __init__(self, fuel_indicator_dict: List):
        self.remaining_range_fuel: Optional[Tuple[int, str]] = None
        self.remaining_range_electric: Optional[Tuple[int, str]] = None
        self.remaining_range_combined: Optional[Tuple[int, str]] = None
        self.remaining_charging_time: float = None
        self.charging_status: str = None
        self.charging_start_time: datetime.datetime = None
        self.charging_end_time: datetime.datetime = None
        self.charging_time_label: str = None

        self._map_to_attributes(fuel_indicator_dict)

    def _map_to_attributes(self, fuel_indicators: List[Dict]) -> None:
        """Parse fuel indicators based on Ids."""
        for indicator in fuel_indicators:
            if (indicator.get("rangeIconId") or indicator.get("infoIconId")) == 59691:  # Combined
                self.remaining_range_combined = self._parse_to_tuple(indicator)
            elif (indicator.get("rangeIconId") or indicator.get("infoIconId")) == 59683:  # Electric
                self.remaining_range_electric = self._parse_to_tuple(indicator)
                self.remaining_range_combined = self.remaining_range_combined or self.remaining_range_electric

                self.charging_time_label = indicator["infoLabel"]
                self.charging_status = indicator["chargingStatusType"]

                if indicator.get("chargingStatusType") in ["CHARGING", "PLUGGED_IN"]:
                    self._parse_charging_timestamp(indicator)

            elif (indicator.get("rangeIconId") or indicator.get("infoIconId")) == 59681:  # Fuel
                self.remaining_range_fuel = self._parse_to_tuple(indicator)
                self.remaining_range_combined = self.remaining_range_combined or self.remaining_range_fuel

    def _parse_charging_timestamp(self, indicator: Dict) -> None:
        """Parse charging end time string to timestamp."""
        charging_start_time: datetime.datetime = None
        charging_end_time: datetime.datetime = None
        remaining_charging_time: int = None

        # Only calculate charging end time if infolabel is like '100% at ~11:04am'
        # Other options: 'Charging', 'Starts at ~09:00am' (but not handled here)

        time_str = indicator["infoLabel"].split("~")[-1].strip()
        try:
            time_parsed = datetime.datetime.strptime(time_str, "%I:%M %p")

            current_time = datetime.datetime.now()
            datetime_parsed = time_parsed.replace(
                year=current_time.year,
                month=current_time.month,
                day=current_time.day
            )
            if datetime_parsed < current_time:
                datetime_parsed = datetime_parsed + datetime.timedelta(days=1)

            if indicator["chargingStatusType"] == "CHARGING":
                charging_end_time = datetime_parsed
                remaining_charging_time = (charging_end_time - current_time).seconds
            elif indicator["chargingStatusType"] == "PLUGGED_IN":
                charging_start_time = datetime_parsed
        except ValueError:
            _LOGGER.error(
                "Error parsing charging end time '%s' out of '%s'",
                time_str,
                indicator["infoLabel"]
            )

        self.charging_end_time = charging_end_time
        self.charging_start_time = charging_start_time
        self.remaining_charging_time = remaining_charging_time

    @staticmethod
    def _parse_to_tuple(fuel_indicator):
        """Parse fuel indicator to standard range tuple."""
        try:
            range_val = int(fuel_indicator["rangeValue"])
        except ValueError:
            return None
        return (range_val, fuel_indicator["rangeUnits"])


def backend_parameter(func):
    """Decorator for parameters reading data from the backend.

    Errors are handled in a default way.
    """
    def _func_wrapper(self: 'VehicleStatus', *args, **kwargs):
        # pylint: disable=protected-access
        if self.properties is None and self.status is None:
            raise ValueError('No data available for vehicle status!')
        try:
            return func(self, *args, **kwargs)
        except KeyError:
            _LOGGER.debug('No data available for attribute %s!', str(func))
            return None
    return _func_wrapper


class VehicleStatus(SerializableBaseClass):  # pylint: disable=too-many-public-methods
    """Models the status of a vehicle."""

    def __init__(self, account: "ConnectedDriveAccount", status_dict: Dict = None):
        """Constructor."""
        self._account = account
        self.status: Dict = {}
        self.properties: Dict = {}
        self._fuel_indicators: FuelIndicator = {}
        self._remote_service_position: Dict = {}

        if status_dict:
            self.update_state(status_dict)

    def update_state(self, status_dict: Dict):
        """Updates the vehicle status."""
        self.status: Dict = status_dict["status"]
        self.properties: Dict = status_dict["properties"]
        self._fuel_indicators = FuelIndicator(status_dict["status"]["fuelIndicators"])

    def set_remote_service_position(self, position_dict: Dict):
        """Store remote service position returned from vehicle finder service."""
        if position_dict.get('errorDetails'):
            error = position_dict["errorDetails"]
            _LOGGER.error("Error retrieving vehicle position. %s: %s", error["title"], error["description"])
            return None
        pos = position_dict["positionData"]["position"]
        pos["timestamp"] = datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc)

        self._remote_service_position = pos
        return None

    @property
    @backend_parameter
    def timestamp(self) -> datetime.datetime:
        """Get the timestamp when the data was recorded."""
        return max(
            parse_datetime(self.properties['lastUpdatedAt']),
            parse_datetime(self.status['lastUpdatedAt'])
        )

    @property
    @backend_parameter
    def gps_position(self) -> Tuple[float, float]:
        """Get the last known position of the vehicle.

        Returns a tuple of (latitude, longitude).
        This only provides data, if the vehicle tracking is enabled!
        """
        if self.is_vehicle_active:
            _LOGGER.info('Vehicle was moving at last update, no position available')
            return (None, None)
        if not self._remote_service_position and "vehicleLocation" not in self.properties:
            _LOGGER.info("No vehicle location data available.")
            return (None, None)

        t_remote = self._remote_service_position.get(
            "timestamp",
            datetime.datetime(1900, 1, 1, tzinfo=datetime.timezone.utc)
        )
        if t_remote > self.timestamp:
            pos = self._remote_service_position
        else:
            pos = self.properties['vehicleLocation']["coordinates"]

        # Convert GCJ02 to WGS84 for positions in China
        if self._account.region == Regions.CHINA:
            pos = pos.copy()
            pos['longitude'], pos['latitude'] = gcj2wgs(gcjLon=pos['longitude'], gcjLat=pos['latitude'])

        return float(pos['latitude']), float(pos['longitude'])

    @property
    @backend_parameter
    def gps_heading(self) -> int:
        """Get the last known heading of the vehicle.

        This only provides data, if the vehicle tracking is enabled!
        """
        if self.is_vehicle_active:
            _LOGGER.info('Vehicle was moving at last update, no position available')
            return None

        if not self._remote_service_position and "vehicleLocation" not in self.properties:
            _LOGGER.info("No vehicle location data available.")
            return None

        t_remote = self._remote_service_position.get(
            "timestamp",
            datetime.datetime(1900, 1, 1, tzinfo=datetime.timezone.utc)
        )
        if t_remote > self.timestamp:
            pos = self._remote_service_position
        else:
            pos = self.properties['vehicleLocation']

        return int(pos['heading'])

    @property
    @backend_parameter
    def is_vehicle_active(self) -> bool:
        """Check if the vehicle is active/moving.

        If the vehicle was active/moving at the time of the last status update, current position is not available.
        """
        return self.properties['inMotion']

    @property
    @backend_parameter
    def mileage(self) -> Tuple[int, str]:
        """Get the mileage of the vehicle.

        Returns a tuple of (value, unit_of_measurement)
        """
        return (
            self.status['currentMileage']['mileage'],
            self.status['currentMileage']['units']
        )

    @property
    @backend_parameter
    def remaining_range_fuel(self) -> Tuple[int, str]:
        """Get the remaining range of the vehicle on fuel.

        Returns a tuple of (value, unit_of_measurement)
        """
        return self._fuel_indicators.remaining_range_fuel or (None, None)

    @property
    @backend_parameter
    def remaining_fuel(self) -> Tuple[int, str]:
        """Get the remaining fuel of the vehicle.

        Returns a tuple of (value, unit_of_measurement)
        """
        return (
            self.properties['fuelLevel']['value'],
            self.properties['fuelLevel']['units'],
        )

    @property
    @backend_parameter
    def fuel_indicator_count(self) -> int:
        """Gets the number of fuel indicators.

        Can be used to identify REX vehicles if driveTrain == ELECTRIC.
        """
        return len(self.status["fuelIndicators"])

    @property
    @backend_parameter
    def lids(self) -> List['Lid']:
        """Get all lids (doors+hatch+trunk) of the car."""
        result = []
        lids = self.properties["doorsAndWindows"]
        result.extend([Lid(k, v) for k, v in lids.items() if k in ["hood", "trunk"] and v != LidState.INVALID.value])
        result.extend([Lid(k, v) for k, v in lids["doors"].items() if v != LidState.INVALID.value])

        return result

    @property
    def open_lids(self) -> List['Lid']:
        """Get all open lids of the car."""
        return [lid for lid in self.lids if not lid.is_closed]

    @property
    def all_lids_closed(self) -> bool:
        """Check if all lids are closed."""
        return len(list(self.open_lids)) == 0

    @property
    @backend_parameter
    def windows(self) -> List['Window']:
        """Get all windows (doors+sun roof) of the car."""
        result = [
            Window(k, v)
            for k, v in self.properties["doorsAndWindows"].get("windows").items()
            if v != LidState.INVALID.value
        ]
        if "moonroof" in self.properties["doorsAndWindows"]:
            result.append(Window("moonroof", self.properties["doorsAndWindows"]["moonroof"]))
        return result

    @property
    def open_windows(self) -> List['Window']:
        """Get all open windows of the car."""
        return [lid for lid in self.windows if not lid.is_closed]

    @property
    def all_windows_closed(self) -> bool:
        """Check if all windows are closed."""
        return len(self.open_windows) == 0

    @property
    @backend_parameter
    def door_lock_state(self) -> LockState:
        """Get state of the door locks."""
        return LockState(self.status['doorsGeneralState'].upper())

    @property
    @backend_parameter
    def last_update_reason(self) -> str:
        """The reason for the last state update"""
        return self.status['timestampMessage']

    @property
    @backend_parameter
    def last_charging_end_result(self) -> str:
        """Get the last charging end result"""
        return None  # Not available in My BMW

    @property
    @backend_parameter
    def connection_status(self) -> str:
        """Get status of the connection"""
        if "chargingState" not in self.properties:
            return None
        return (
            "CONNECTED"
            if self.properties["chargingState"]["isChargerConnected"]
            else "DISCONNECTED"
        )

    @property
    @backend_parameter
    def condition_based_services(self) -> List['ConditionBasedServiceReport']:
        """Get status of the condition based services."""
        return [ConditionBasedServiceReport(s) for s in self.properties['serviceRequired']]

    @property
    def are_all_cbs_ok(self) -> bool:
        """Check if the status of all condition based services is "OK"."""
        for cbs in self.condition_based_services:
            if cbs.state != ConditionBasedServiceStatus.OK:
                return False
        return True

    @property
    @backend_parameter
    def parking_lights(self) -> None:
        """Get status of parking lights.

        :returns None if status is unknown.
        """
        return None  # Not available in My BMW

    @property
    def has_parking_light_state(self) -> bool:
        """Return True if parking light is available."""
        return False  # Not available in My BMW

    @property
    def are_parking_lights_on(self) -> bool:
        """Get status of parking lights.

        :returns None if status is unknown.
        """
        return None  # Not available in My BMW

    @property
    @backend_parameter
    def remaining_range_electric(self) -> Tuple[int, str]:
        """Remaining range on battery, in kilometers."""
        return self._fuel_indicators.remaining_range_electric or (None, None)

    @property
    @backend_parameter
    def remaining_range_total(self) -> Tuple[int, str]:
        """Get the total remaining range of the vehicle in kilometers.

        That is electrical range + fuel range.
        """
        return self._fuel_indicators.remaining_range_combined or (None, None)

    @property
    @backend_parameter
    def max_range_electric(self) -> int:
        """ This can change with driving style and temperature in kilometers."""
        return None  # Not available in My BMW

    @property
    @backend_parameter
    def charging_status(self) -> ChargingState:
        """Charging state of the vehicle."""
        if "chargingState" not in self.properties:
            return None
        return ChargingState(self._fuel_indicators.charging_status)

    @property
    @backend_parameter
    def charging_time_remaining(self) -> float:
        """Get the remaining charging duration."""
        return round((self._fuel_indicators.remaining_charging_time or 0) / 60.0 / 60.0, 2)

    @property
    @backend_parameter
    def charging_start_time(self) -> datetime.datetime:
        """Get the charging finish time."""
        if self._fuel_indicators.charging_start_time:
            return self._fuel_indicators.charging_start_time.replace(tzinfo=self._account.timezone())
        return None

    @property
    @backend_parameter
    def charging_end_time(self) -> datetime.datetime:
        """Get the charging finish time."""
        if self._fuel_indicators.charging_end_time:
            return self._fuel_indicators.charging_end_time.replace(tzinfo=self._account.timezone())
        return None

    @property
    @backend_parameter
    def charging_time_label(self) -> datetime.datetime:
        """Get the remaining charging time as provided by the API."""
        return self._fuel_indicators.charging_time_label

    @property
    @backend_parameter
    def charging_level_hv(self) -> int:
        """State of charge of the high voltage battery in percent."""
        return int(self.properties["electricRangeAndStatus"]["chargePercentage"])

    @property
    @backend_parameter
    def fuel_percent(self) -> int:
        """State of fuel in percent."""
        return int(self.properties['fuelPercentage']["value"])

    @property
    @backend_parameter
    def check_control_messages(self) -> List[CheckControlMessage]:
        """List of check control messages."""
        messages = self.status.get('checkControlMessages', [])
        return [CheckControlMessage(m) for m in messages if m["state"] != "OK"]

    @property
    @backend_parameter
    def has_check_control_messages(self) -> bool:
        """Return true if any check control message is present."""
        return len(self.check_control_messages) > 0


class Lid:  # pylint: disable=too-few-public-methods
    """A lid of the vehicle.

    Lids are: Doors + Trunk + Hatch
    """

    def __init__(self, name: str, state: str):
        #: name of the lid
        self.name = name
        self.state = LidState(state)

    @property
    def is_closed(self) -> bool:
        """Check if the lid is closed."""
        return self.state == LidState.CLOSED


class Window(Lid):  # pylint: disable=too-few-public-methods,no-member
    """A window of the vehicle.

    A window can be a normal window of the car or the sun roof.
    """


class ConditionBasedServiceReport:  # pylint: disable=too-few-public-methods
    """Entry in the list of condition based services."""

    def __init__(self, cbs_data: dict):

        #: date when the service is due
        self.due_date = parse_datetime(cbs_data.get('dateTime'))

        #: status of the service
        self.state = ConditionBasedServiceStatus(cbs_data['status'])

        #: service type
        self.service_type = cbs_data['type']

        #: distance when the service is due
        self.due_distance = None
        if 'distance' in cbs_data:
            self.due_distance = (cbs_data["distance"]['value'], cbs_data["distance"]['units'])

        #: description of the required service
        self.description = None  # Could be retrieved from status.requiredServices if needed



"""Tests for ConnectedDriveAccount."""
import datetime
import logging
import os
import time
import unittest
from unittest.mock import Mock

import time_machine

from bimmer_connected.country_selector import get_region_from_name, valid_regions
from bimmer_connected.utils import get_class_property_names, parse_datetime, to_json

from . import RESPONSE_DIR, VIN_G21
from .test_account import get_mocked_account


class TestUtils(unittest.TestCase):
    """Tests for utils."""

    def test_drive_train(self):
        """Tests available attribute."""
        vehicle = get_mocked_account().get_vehicle(VIN_G21)
        self.assertEqual(
            [
                "available_attributes",
                "available_state_services",
                "brand",
                "charging_profile",
                "drive_train",
                "drive_train_attributes",
                "has_hv_battery",
                "has_internal_combustion_engine",
                "has_range_extender",
                "has_weekly_planner_service",
                "is_vehicle_tracking_enabled",
                "lsc_type",
                "name",
            ],
            get_class_property_names(vehicle),
        )

    @time_machine.travel("2011-11-28 21:28:59 +0000", tick=False)
    def test_to_json(self):
        """Test serialization to JSON."""
        # Force UTC
        os.environ["TZ"] = "UTC"
        time.tzset()

        account = get_mocked_account()
        account.timezone = Mock(return_value=datetime.timezone.utc)
        vehicle = account.get_vehicle(VIN_G21)

        # Unset UTC after vehicle has been loaded
        del os.environ["TZ"]
        time.tzset()

        with open(RESPONSE_DIR / "G21" / "json_export.json", "rb") as file:
            expected = file.read().decode("UTF-8")
        self.assertEqual(expected, to_json(vehicle, indent=4))

    def test_parse_datetime(self):
        """Test datetime parser."""

        dt_with_milliseconds = datetime.datetime(2021, 11, 12, 13, 14, 15, 567000, tzinfo=datetime.timezone.utc)
        dt_without_milliseconds = datetime.datetime(2021, 11, 12, 13, 14, 15, tzinfo=datetime.timezone.utc)

        self.assertEqual(dt_with_milliseconds, parse_datetime("2021-11-12T13:14:15.567Z"))

        self.assertEqual(dt_without_milliseconds, parse_datetime("2021-11-12T13:14:15Z"))

        with self.assertLogs(level=logging.ERROR):
            self.assertIsNone(parse_datetime("2021-14-12T13:14:15Z"))


class TestCountrySelector(unittest.TestCase):
    """Tests for Country Selector."""

    def test_valid_regions(self):
        """Test valid regions."""
        self.assertListEqual(["north_america", "china", "rest_of_world"], valid_regions())

    def test_unknown_region(self):
        """Test unkown region."""
        with self.assertRaises(ValueError):
            get_region_from_name("unkown")



"""
Convert chinese mapping systems to WGS84. 

Copied from https://github.com/sshuair/coord-convert/blob/master/coord_convert/transform.py 
as other dependencies except actual conversion were not needed.

Source: transform.py of https://pypi.org/project/coord-convert/
Author: sshuair
License: MIT License
"""

# pylint: skip-file
# flake8: noqa

# -*- coding: utf-8 -*-
from math import sin, cos, sqrt, fabs, atan2
from math import pi as PI


# define ellipsoid
a = 6378245.0
f = 1 / 298.3
b = a * (1 - f)
ee = 1 - (b * b) / (a * a)


def outOfChina(lng, lat):
    """check weather lng and lat out of china
    
    Arguments:
        lng {float} -- longitude
        lat {float} -- latitude
    
    Returns:
        Bollen -- True or False
    """
    return not (72.004 <= lng <= 137.8347 and 0.8293 <= lat <= 55.8271)


def transformLat(x, y):
    ret = -100.0 + 2.0 * x + 3.0 * y + 0.2 * y * y + 0.1 * x * y + 0.2 * sqrt(fabs(x))
    ret = ret + (20.0 * sin(6.0 * x * PI) + 20.0 * sin(2.0 * x * PI)) * 2.0 / 3.0
    ret = ret + (20.0 * sin(y * PI) + 40.0 * sin(y / 3.0 * PI)) * 2.0 / 3.0
    ret = ret + (160.0 * sin(y / 12.0 * PI) + 320.0 * sin(y * PI / 30.0)) * 2.0 / 3.0
    return ret


def transformLon(x, y):
    ret = 300.0 + x + 2.0 * y + 0.1 * x * x +  0.1 * x * y + 0.1 * sqrt(fabs(x))
    ret = ret + (20.0 * sin(6.0 * x * PI) + 20.0 * sin(2.0 * x * PI)) * 2.0 / 3.0
    ret = ret + (20.0 * sin(x * PI) + 40.0 * sin(x / 3.0 * PI)) * 2.0 / 3.0
    ret = ret + (150.0 * sin(x / 12.0 * PI) + 300.0 * sin(x * PI / 30.0)) * 2.0 / 3.0
    return ret


def wgs2gcj(wgsLon, wgsLat):
    """wgs coord to gcj
    
    Arguments:
        wgsLon {float} -- lon
        wgsLat {float} -- lat
    
    Returns:
        tuple -- gcj coords
    """

    if outOfChina(wgsLon, wgsLat):
        return wgsLon, wgsLat
    dLat = transformLat(wgsLon - 105.0, wgsLat - 35.0)
    dLon = transformLon(wgsLon - 105.0, wgsLat - 35.0)
    radLat = wgsLat / 180.0 * PI
    magic = sin(radLat)
    magic = 1 - ee * magic * magic
    sqrtMagic = sqrt(magic)
    dLat = (dLat * 180.0) / ((a * (1 - ee)) / (magic * sqrtMagic) * PI)
    dLon = (dLon * 180.0) / (a / sqrtMagic * cos(radLat) * PI)
    gcjLat = wgsLat + dLat
    gcjLon = wgsLon + dLon
    return (gcjLon, gcjLat)


def gcj2wgs(gcjLon, gcjLat):
    g0 = (gcjLon, gcjLat)
    w0 = g0
    g1 = wgs2gcj(w0[0], w0[1])
    # w1 = w0 - (g1 - g0)
    w1 = tuple(map(lambda x: x[0]-(x[1]-x[2]), zip(w0,g1,g0)))
    # delta = w1 - w0
    delta = tuple(map(lambda x: x[0] - x[1], zip(w1, w0)))
    while (abs(delta[0]) >= 1e-6 or abs(delta[1]) >= 1e-6):
        w0 = w1
        g1 = wgs2gcj(w0[0], w0[1])
        # w1 = w0 - (g1 - g0)
        w1 = tuple(map(lambda x: x[0]-(x[1]-x[2]), zip(w0,g1,g0)))
        # delta = w1 - w0
        delta = tuple(map(lambda x: x[0] - x[1], zip(w1, w0)))
    return w1


def gcj2bd(gcjLon, gcjLat):
    z = sqrt(gcjLon * gcjLon + gcjLat * gcjLat) + 0.00002 * sin(gcjLat * PI * 3000.0 / 180.0)
    theta = atan2(gcjLat, gcjLon) + 0.000003 * cos(gcjLon * PI * 3000.0 / 180.0)
    bdLon = z * cos(theta) + 0.0065
    bdLat = z * sin(theta) + 0.006
    return (bdLon, bdLat)


def bd2gcj(bdLon, bdLat):
    x = bdLon - 0.0065
    y = bdLat - 0.006
    z = sqrt(x * x + y * y) - 0.00002 * sin(y * PI * 3000.0 / 180.0)
    theta = atan2(y, x) - 0.000003 * cos(x * PI * 3000.0 / 180.0)
    gcjLon = z * cos(theta)
    gcjLat = z * sin(theta)
    return (gcjLon, gcjLat)


def wgs2bd(wgsLon, wgsLat):
    gcj = wgs2gcj(wgsLon, wgsLat)
    return gcj2bd(gcj[0], gcj[1])


def bd2wgs(bdLon, bdLat):
    gcj = bd2gcj(bdLon, bdLat)
    return gcj2wgs(gcj[0], gcj[1])


class Transform():

    def transformLat(self, x, y):
        return transformLat(x, y)

    def transformLon(self, x, y):
        return transformLon(x, y)

    def wgs2gcj(self, wgsLon, wgsLat):
        return wgs2gcj(wgsLon, wgsLat)

    def gcj2wgs(self, gcjLon, gcjLat):
        return gcj2wgs(gcjLon, gcjLat)

    def gcj2bd(self, gcjLon, gcjLat):
        return gcj2bd(gcjLon, gcjLat)

    def bd2gcj(self, bdLon, bdLat):
        return bd2gcj(bdLon, bdLat)

    def wgs2bd(self, wgsLon, wgsLat):
        return wgs2bd(wgsLon, wgsLat)

    def bd2wgs(self, bdLon, bdLat):
        return bd2wgs(bdLon, bdLat)



"""Test for remote_services."""
import logging
import re

from unittest import TestCase, mock

import time_machine
import requests_mock
from requests.exceptions import HTTPError

from bimmer_connected import remote_services
from bimmer_connected.remote_services import ExecutionState, RemoteServiceStatus


from . import RESPONSE_DIR, VIN_F45, load_response
from .test_account import get_base_adapter, get_mocked_account

_RESPONSE_INITIATED = RESPONSE_DIR / "remote_services" / "eadrax_service_initiated.json"
_RESPONSE_PENDING = RESPONSE_DIR / "remote_services" / "eadrax_service_pending.json"
_RESPONSE_DELIVERED = RESPONSE_DIR / "remote_services" / "eadrax_service_delivered.json"
_RESPONSE_EXECUTED = RESPONSE_DIR / "remote_services" / "eadrax_service_executed.json"
_RESPONSE_ERROR = RESPONSE_DIR / "remote_services" / "eadrax_service_error.json"
_RESPONSE_EVENTPOSITION = RESPONSE_DIR / "remote_services" / "eadrax_service_eventposition.json"


POI_DATA = {
    "lat": 37.4028943,
    "lon": -121.9700289,
    "name": "49ers",
    "additional_info": "Hi Sam",
    "street": "4949 Marie P DeBartolo Way",
    "city": "Santa Clara",
    "postal_code": "CA 95054",
    "country": "United States",
    "website": "https://www.49ers.com/",
    "phone_numbers": ["+1 408-562-4949"],
}


def get_remote_services_adapter():
    """Returns mocked adapter for auth."""
    adapter = get_base_adapter()
    adapter.register_uri(
        "POST",
        re.compile(r"/eadrax-vrccs/v2/presentation/remote-commands/.+/.+$"),
        json=load_response(_RESPONSE_INITIATED),
    )
    adapter.register_uri(
        "POST",
        re.compile(r"/eadrax-vrccs/v2/presentation/remote-commands/eventStatus\?eventId=.+$"),
        [
            {"json": load_response(_RESPONSE_PENDING)},
            {"json": load_response(_RESPONSE_DELIVERED)},
            {"json": load_response(_RESPONSE_EXECUTED)},
        ],
    )
    adapter.register_uri("POST", "/eadrax-dcs/v1/send-to-car/send-to-car", status_code=201)
    adapter.register_uri(
        "POST",
        re.compile(r"/eadrax-vrccs/v2/presentation/remote-commands/eventPosition\?eventId=.+$"),
        json=load_response(_RESPONSE_EVENTPOSITION),
    )
    return adapter


class TestRemoteServices(TestCase):
    """Test for remote_services."""

    # pylint: disable=protected-access

    def test_states(self):
        """Test parsing the different response types."""
        rss = RemoteServiceStatus(load_response(_RESPONSE_PENDING))
        self.assertEqual(ExecutionState.PENDING, rss.state)

        rss = RemoteServiceStatus(load_response(_RESPONSE_DELIVERED))
        self.assertEqual(ExecutionState.DELIVERED, rss.state)

        rss = RemoteServiceStatus(load_response(_RESPONSE_EXECUTED))
        self.assertEqual(ExecutionState.EXECUTED, rss.state)

    def test_trigger_remote_services(self):
        """Test executing a remote light flash."""
        remote_services._POLLING_CYCLE = 0
        remote_services._UPDATE_AFTER_REMOTE_SERVICE_DELAY = 0

        services = [
            ("LIGHT_FLASH", "trigger_remote_light_flash", False),
            ("DOOR_LOCK", "trigger_remote_door_lock", True),
            ("DOOR_UNLOCK", "trigger_remote_door_unlock", True),
            ("CLIMATE_NOW", "trigger_remote_air_conditioning", True),
            ("CLIMATE_STOP", "trigger_remote_air_conditioning_stop", True),
            ("VEHICLE_FINDER", "trigger_remote_vehicle_finder", False),
            ("HORN_BLOW", "trigger_remote_horn", False),
            ("CHARGE_NOW", "trigger_charge_now", True),
            ("SEND_POI", "trigger_send_poi", False),
        ]

        with requests_mock.Mocker(adapter=get_remote_services_adapter()):
            for service, call, triggers_update in services:
                account = get_mocked_account()
                mock_listener = mock.Mock(return_value=None)
                account.add_update_listener(mock_listener)
                vehicle = account.get_vehicle(VIN_F45)

                if service == "SEND_POI":
                    response = getattr(vehicle.remote_services, call)(POI_DATA)
                else:
                    response = getattr(vehicle.remote_services, call)()
                    self.assertEqual(ExecutionState.EXECUTED, response.state)

                    if triggers_update:
                        mock_listener.assert_called_once_with()
                    else:
                        mock_listener.assert_not_called()

    def test_get_remote_service_status(self):
        """Test get_remove_service_status method."""
        remote_services._POLLING_CYCLE = 0
        remote_services._UPDATE_AFTER_REMOTE_SERVICE_DELAY = 0

        account = get_mocked_account()
        vehicle = account.get_vehicle(VIN_F45)

        with requests_mock.Mocker() as mocker:
            mocker.post(
                "/eadrax-vrccs/v2/presentation/remote-commands/eventStatus?eventId=None",
                [
                    dict(status_code=500, json=[]),
                    dict(status_code=200, text="You can't parse this..."),
                    dict(status_code=200, json=load_response(_RESPONSE_ERROR)),
                ],
            )

            with self.assertRaises(HTTPError):
                vehicle.remote_services._get_remote_service_status(remote_services._Services.REMOTE_LIGHT_FLASH)
            with self.assertRaises(ValueError):
                vehicle.remote_services._get_remote_service_status(remote_services._Services.REMOTE_LIGHT_FLASH)
            with self.assertRaises(Exception):
                vehicle.remote_services._block_until_done(event_id=remote_services._Services.REMOTE_LIGHT_FLASH)

    def test_get_remote_position(self):
        """Test getting position from remote service."""
        remote_services._POLLING_CYCLE = 0
        remote_services._UPDATE_AFTER_REMOTE_SERVICE_DELAY = 0

        with requests_mock.Mocker(adapter=get_remote_services_adapter()):
            account = get_mocked_account()
            account.set_observer_position(1.0, 0.0)
            vehicle = account.get_vehicle(VIN_F45)
            status = vehicle.status

            # Check original position
            self.assertTupleEqual((12.3456, 34.5678), status.gps_position)
            self.assertAlmostEqual(123, status.gps_heading)

            # Check updated position
            vehicle.remote_services.trigger_remote_vehicle_finder()
            self.assertTupleEqual((123.456, 34.5678), status.gps_position)
            self.assertAlmostEqual(121, status.gps_heading)

            # Position should still be from vehicle finder after status update
            account._get_vehicles()
            self.assertTupleEqual((123.456, 34.5678), status.gps_position)
            self.assertAlmostEqual(121, status.gps_heading)

    def test_get_remote_position_fail_without_observer(self):
        """Test getting position from remote service."""
        remote_services._POLLING_CYCLE = 0
        remote_services._UPDATE_AFTER_REMOTE_SERVICE_DELAY = 0

        with requests_mock.Mocker(adapter=get_remote_services_adapter()):
            account = get_mocked_account()
            vehicle = account.get_vehicle(VIN_F45)

            with self.assertLogs(level=logging.ERROR):
                vehicle.remote_services.trigger_remote_vehicle_finder()

    @time_machine.travel("2020-01-01", tick=False)
    def test_get_remote_position_too_old(self):
        """Test remote service position being ignored as vehicle status is newer."""
        remote_services._POLLING_CYCLE = 0
        remote_services._UPDATE_AFTER_REMOTE_SERVICE_DELAY = 0

        with requests_mock.Mocker(adapter=get_remote_services_adapter()):
            account = get_mocked_account()
            vehicle = account.get_vehicle(VIN_F45)
            status = vehicle.status

            vehicle.remote_services.trigger_remote_vehicle_finder()

            self.assertTupleEqual((12.3456, 34.5678), status.gps_position)
            self.assertAlmostEqual(123, status.gps_heading)

    def test_poi(self):
        """Test get_remove_service_status method."""
        remote_services._POLLING_CYCLE = 0
        remote_services._UPDATE_AFTER_REMOTE_SERVICE_DELAY = 0

        account = get_mocked_account()
        vehicle = account.get_vehicle(VIN_F45)

        with requests_mock.Mocker(adapter=get_remote_services_adapter()):
            with self.assertRaises(TypeError):
                vehicle.remote_services.trigger_send_poi({"lat": 12.34})



"""Mock for Connected Drive Backend."""

import json
from pathlib import Path
from typing import Union

# from requests.exceptions import HTTPError

from bimmer_connected.country_selector import Regions

RESPONSE_DIR = Path(__file__).parent / "responses"

TEST_USERNAME = "some_user"
TEST_PASSWORD = "my_secret"
TEST_REGION = Regions.REST_OF_WORLD
TEST_REGION_STRING = "rest_of_world"

VIN_F11 = "some_vin_F11"
VIN_F31 = "some_vin_F31"
VIN_F35 = "some_vin_F35"
VIN_F44 = "some_vin_F44"
VIN_F45 = "some_vin_F45"
VIN_F48 = "some_vin_F48"
VIN_G01 = "some_vin_G01"
VIN_G05 = "some_vin_G05"
VIN_G08 = "some_vin_G08"
VIN_G21 = "some_vin_G21"
VIN_G30 = "some_vin_G30"
VIN_I01_NOREX = "some_vin_I01_NOREX"
VIN_I01_REX = "some_vin_I01_REX"

ALL_FINGERPRINTS = []


def get_fingerprint_count() -> int:
    """Returns number of loaded vehicles."""
    return len(ALL_FINGERPRINTS)


def load_response(path: Union[Path, str]) -> object:
    """Load a stored response."""
    with open(path, "rb") as file:
        if Path(path).suffix == ".json":
            return json.load(file)
        return file.read().decode("UTF-8")


for fingerprint in RESPONSE_DIR.rglob("vehicles_v2_*_0.json"):
    ALL_FINGERPRINTS.extend(load_response(fingerprint))


# # VehicleState has different names than the json file. So we need to map some of the
# # parameters.
# ATTRIBUTE_MAPPING = {
#     'remainingFuel': 'remaining_fuel',
#     'position': 'gps_position',
#     'cbsData': 'condition_based_services',
#     'checkControlMessages': 'check_control_messages',
#     'doorLockState': 'door_lock_state',
#     'updateReason': 'last_update_reason',
#     'chargingLevelHv': 'charging_level_hv',
#     'chargingStatus': 'charging_status',
#     'maxRangeElectric': 'max_range_electric',
#     'remainingRangeElectric': 'remaining_range_electric',
#     'parkingLight': 'parking_lights',
#     'remainingRangeFuel': 'remaining_range_fuel',
#     'updateTime': 'timestamp',
#     'chargingTimeRemaining': 'charging_time_remaining'
# }

# # these are additional attributes in the API, not available in the status.json
# ADDITIONAL_ATTRIBUTES = [
#     'remaining_range_total',    # added by bimmer_connected
#     'charging_time_remaining',  # only present while charging
#     'sunroof',                  # not available in all vehicles
#     'lids',                     # required for existing Home Assistant binary sensors
#     'windows',                  # required for existing Home Assistant binary sensors
#     'lights_parking',           # required for existing Home Assistant binary sensors
#     'steering',                 # para not available in all vehicles
# ]

# # there attributes are not (yet) implemented
# MISSING_ATTRIBUTES = [
#     'remainingRangeFuelMls',      # we're not using miles
#     'remainingRangeElectricMls',  # we're not using miles
#     'maxRangeElectricMls',        # we're not using miles
#     'chargingTimeRemaining',      # only present while charging
#     'sunroof',                    # not available in all vehicles
#     'lights_parking',             # required for existing Home Assistant binary sensors
#     'steering',                   # para not available in all vehicles
#     'vehicleCountry',             # para not available in all vehicles
# ]

# AVAILABLE_STATES_MAPPING = {
#     "statisticsAvailable": {True: ["LAST_TRIP", "ALL_TRIPS"]},
#     "chargingControl": {"WEEKLY_PLANNER": ["CHARGING_PROFILE"]},
#     "lastDestinations": {"SUPPORTED": ["DESTINATIONS"]},
#     "rangeMap": {"RANGE_CIRCLE": ["RANGEMAP"]}
# }

# POI_DATA = {
#     "lat": 37.4028943,
#     "lon": -121.9700289,
#     "name": "49ers",
#     "additional_info": "Hi Sam",
#     "street": "4949 Marie P DeBartolo Way",
#     "city": "Santa Clara",
#     "postal_code": "CA 95054",
#     "country": "United States",
#     "website": "https://www.49ers.com/",
#     "phone_numbers": ["+1 408-562-4949"]
# }

# POI_REQUEST = {
#     "min": ("data=%7B%22poi%22%3A+%7B%22lat%22%3A+37.4028943%2C+%22lon%22%3A+-121.9700289"
#             "%2C+%22additionalInfo%22%3A+%22Sent+with+%5Cu2665+by+bimmer_connected%22%7D%7D"),
#     "all": ("data=%7B%22poi%22%3A+%7B%22lat%22%3A+37.4028943%2C+%22lon%22%3A+-121.9700289"
#             "%2C+%22name%22%3A+%2249ers%22%2C+%22additionalInfo%22%3A+%22Hi+Sam%22"
#             "%2C+%22street%22%3A+%224949+Marie+P+DeBartolo+Way%22%2C+%22city%22%3A+%22Santa+Clara%22"
#             "%2C+%22postalCode%22%3A+%22CA+95054%22%2C+%22country%22%3A+%22United+States%22"
#             "%2C+%22website%22%3A+%22https%3A%2F%2Fwww.49ers.com%2F%22"
#             "%2C+%22phoneNumbers%22%3A+%5B%22%2B1+408-562-4949%22%5D%7D%7D")
# }

# MESSAGE_DATA = {"subject": "This is a subject...", "text": "This is a message!"}

# MESSAGE_REQUEST = {
#     "min": ("data=%7B%22poi%22%3A+%7B%22additionalInfo%22%3A+%22This+is+a+message%21%22%7D%7D"),
#     "all": ("data=%7B%22poi%22%3A+%7B%22name%22%3A+%22This+is+a+subject...%22%2C+%22additionalInfo"
#             "%22%3A+%22This+is+a+message%21%22%7D%7D")
# }


# def load_response_json(filename: str) -> dict:
#     """load a stored response from a file"""
#     with open(os.path.join(RESPONSE_DIR, filename), encoding='UTF-8') as json_file:
#         return json.load(json_file)


# class BackendMock:
#     """Mock for Connected Drive Backend."""

#     # pylint: disable=too-many-arguments

#     def __init__(self) -> None:
#         """Constructor."""
#         self.last_request = []
#         self.responses = [
#             MockResponse('https://.+/gcdm/.*/?oauth/authenticate',
#                          headers=_AUTH_RESPONSE_HEADERS,
#                          data_files=['auth/authorization_response.json'],
#                          status_code=200),
#             MockResponse('https://.+/gcdm/.*/?oauth/token',
#                          headers=_AUTH_RESPONSE_HEADERS,
#                          data_files=['auth/auth_token.json'],
#                          status_code=200),
#             MockResponse('https://.+/webapi/v1/user/vehicles$',
#                          data_files=['vehicles.json']),
#         ]

#     def get(self, url: str, headers: dict = None, data: str = None, allow_redirects: bool = None, params=None) \
#             -> 'MockResponse':
#         """Mock for requests.get function."""
#         self.last_request.append(MockRequest(url, headers, data, request_type='GET', allow_redirects=allow_redirects,
#                                              params=params))
#         return self._find_response(url)

#     def post(self, url: str, headers: dict = None, data: str = None, allow_redirects: bool = None, params=None) \
#             -> 'MockResponse':
#         """Mock for requests.post function."""
#         self.last_request.append(MockRequest(url, headers, data, request_type='GET', allow_redirects=allow_redirects,
#                                              params=params))
#         return self._find_response(url)

#     def add_response(self, regex: str, data: str = None, data_files: List[str] = None,
#                      headers: dict = None, status_code=200) -> None:
#         """Add a response to the backend."""
#         self.responses.append(MockResponse(regex, data, data_files, headers, status_code))

#     def _find_response(self, url) -> 'MockResponse':
#         """Find a proper response for a requested url."""
#         for response in self.responses:
#             if response.regex.search(url):
#                 response.step_responses()
#                 return response
#         return MockResponse(regex='', data='unknown url: {}'.format(url), status_code=404)

#     def setup_default_vehicles(self) -> None:
#         """Setup the vehicle configuration in a mock backend."""
#         for vin, path in TEST_VEHICLE_DATA.items():
#             self.add_response('https://.+/webapi/v1/user/vehicles/{vin}/status$'.format(vin=vin),
#                               data_files=['{path}/status.json'.format(path=path)])

#     def Session(self) -> 'BackendMock':  # pylint: disable=invalid-name
#         """Returns itself as a requests.Session style object"""
#         return self


# class MockRequest:  # pylint: disable=too-few-public-methods
#     """Stores the attributes of a request."""

#     # pylint: disable=too-many-arguments

#     def __init__(self, url, headers, data, request_type=None, allow_redirects=None, params=None) -> None:
#         self.url = url
#         self.headers = headers
#         self.headers["Host"] = urllib.parse.urlparse(url).netloc
#         self.data = data
#         self.request_type = request_type
#         self.allow_redirects = allow_redirects
#         self.params = params


# class MockResponse:
#     """Mocks requests.response."""

#     # pylint: disable=too-many-arguments

#     def __init__(self, regex: str, data: str = None, data_files: List[str] = None, headers: dict = None,
#                  status_code: int = 200) -> None:
#         """Constructor."""
#         self.regex = re.compile(regex)
#         self.status_code = status_code
#         self.headers = headers
#         self._usage_count = 0
#         if self.headers is None:
#             self.headers = {}

#         if data_files is not None:
#             self._data = []
#             for data_file in data_files:
#                 with open(os.path.join(RESPONSE_DIR, data_file), encoding='UTF-8') as response:
#                     self._data.append(response.read())
#         else:
#             self._data = [data]

#     def json(self) -> dict:
#         """Parse the text of the response as a jsons string."""
#         return json.loads(self.text)

#     def raise_for_status(self) -> bool:
#         """Simulate requests' raise_for_status and raise HTTPError if non-ok status code."""
#         http_error_msg = ''
#         if 400 <= self.status_code < 500:
#             http_error_msg = '%s Client Error' % (self.status_code)
#         elif 500 <= self.status_code < 600:
#             http_error_msg = '%s Server Error' % (self.status_code)
#         if http_error_msg:
#             raise HTTPError(http_error_msg, response=self)

#     @property
#     def text(self) -> str:
#         """Get the raw data from the response."""
#         return self._data[self._usage_count-1]

#     @property
#     def next(self) -> str:
#         """Get the next url (i.e. forwarded URL) for a response. Only used for authentication."""
#         class AttributeDict(dict):
#             """Simulate attributes from a dict."""
#             __slots__ = ()
#             __getattr__ = dict.__getitem__
#             __setattr__ = dict.__setitem__
#         return AttributeDict(
#             {
#                 "path_url": "/?code=some_login_code&client_id=31c357a0-7a1d-4590-aa99-33b97244d048&nonce=login_nonce"
#             }
#         )

#     def step_responses(self):
#         """Step through the list of responses.

#         The last response will be repeated forever.
#         """
#         self._usage_count = min(len(self._data), self._usage_count+1)



#!/usr/bin/python3
"""Simple executable to demonstrate and test the usage of the library."""

import argparse
import logging
import time
import sys
from datetime import datetime

from pathlib import Path

import requests

from bimmer_connected.account import ConnectedDriveAccount
from bimmer_connected.country_selector import get_region_from_name, valid_regions, get_server_url
from bimmer_connected.vehicle import VehicleViewDirection, HV_BATTERY_DRIVE_TRAINS
from bimmer_connected.utils import to_json

TEXT_VIN = 'Vehicle Identification Number'


def main_parser() -> argparse.ArgumentParser:
    """Creates the ArgumentParser with all relevant subparsers."""
    logging.basicConfig(level=logging.DEBUG)

    parser = argparse.ArgumentParser(description='A simple executable to use and test the library.')
    subparsers = parser.add_subparsers(dest='cmd')
    subparsers.required = True

    status_parser = subparsers.add_parser('status', description='Get the current status of the vehicle.')
    status_parser.add_argument('-j', '--json',
                               help='Output as JSON only. Removes all other output.',
                               action='store_true'
                               )
    _add_default_arguments(status_parser)
    _add_position_arguments(status_parser)

    fingerprint_parser = subparsers.add_parser('fingerprint', description='Save a vehicle fingerprint.')
    _add_default_arguments(fingerprint_parser)
    _add_position_arguments(fingerprint_parser)
    fingerprint_parser.set_defaults(func=fingerprint)

    flash_parser = subparsers.add_parser('lightflash', description='Flash the vehicle lights.')
    _add_default_arguments(flash_parser)
    flash_parser.add_argument('vin', help=TEXT_VIN)
    flash_parser.set_defaults(func=light_flash)

    finder_parser = subparsers.add_parser('vehiclefinder', description='Update the vehicle GPS location.')
    _add_default_arguments(finder_parser)
    finder_parser.add_argument('vin', help=TEXT_VIN)
    _add_position_arguments(finder_parser)
    finder_parser.set_defaults(func=vehicle_finder)

    image_parser = subparsers.add_parser('image', description='Download a vehicle image.')
    _add_default_arguments(image_parser)
    image_parser.add_argument('vin', help=TEXT_VIN)
    image_parser.set_defaults(func=image)

    sendpoi_parser = subparsers.add_parser('sendpoi', description='Send a point of interest to the vehicle.')
    _add_default_arguments(sendpoi_parser)
    sendpoi_parser.add_argument('vin', help=TEXT_VIN)
    sendpoi_parser.add_argument('latitude', help='Latitude of the POI', type=float)
    sendpoi_parser.add_argument('longitude', help='Longitude of the POI', type=float)
    sendpoi_parser.add_argument('--name', help='(optional, display only) Name of the POI', nargs='?', default=None)
    sendpoi_parser.add_argument('--street', help='(optional, display only) Street & House No. of the POI',
                                nargs='?', default=None)
    sendpoi_parser.add_argument('--city', help='(optional, display only) City of the POI', nargs='?', default=None)
    sendpoi_parser.add_argument('--postalcode', help='(optional, display only) Postal code of the POI',
                                nargs='?', default=None)
    sendpoi_parser.add_argument('--country', help='(optional, display only) Country of the POI',
                                nargs='?', default=None)
    sendpoi_parser.set_defaults(func=send_poi)

    sendpoi_from_address_parser = subparsers.add_parser('sendpoi_from_address',
                                                        description=('Send a point of interest parsed from a'
                                                                     ' street address to the vehicle.'))
    _add_default_arguments(sendpoi_from_address_parser)
    sendpoi_from_address_parser.add_argument('vin', help=TEXT_VIN)
    sendpoi_from_address_parser.add_argument('-n', '--name', help='(optional, display only) Name of the POI',
                                             nargs='?', default=None)
    sendpoi_from_address_parser.add_argument('-a', '--address', nargs='+',
                                             help="Address (e.g. 'Street 17, city, zip, country')")
    sendpoi_from_address_parser.set_defaults(func=send_poi_from_address)

    message_parser = subparsers.add_parser('sendmessage', description='Send a text message to the vehicle.')
    _add_default_arguments(message_parser)
    message_parser.add_argument('vin', help=TEXT_VIN)
    message_parser.add_argument('text', help='Text to be sent.')
    message_parser.add_argument('subject', help='(optional) Message subject', nargs='?')
    message_parser.set_defaults(func=send_message)

    return parser


def get_status(args) -> None:
    """Get the vehicle status."""
    if args.json:
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)

    account = ConnectedDriveAccount(args.username, args.password, get_region_from_name(args.region))
    if args.lat and args.lng:
        for vehicle in account.vehicles:
            vehicle.set_observer_position(args.lat, args.lng)
    account.update_vehicle_states()

    if args.json:
        print(to_json(account.vehicles))
    else:
        print('Found {} vehicles: {}'.format(
            len(account.vehicles),
            ','.join([v.name for v in account.vehicles])))

        for vehicle in account.vehicles:
            print('VIN: {}'.format(vehicle.vin))
            print('Mileage: {}'.format(vehicle.status.mileage))
            print('Vehicle data:')
            print(to_json(vehicle, indent=4))


def fingerprint(args) -> None:
    """Save the vehicle fingerprint."""
    time_dir = Path.home() / 'vehicle_fingerprint' / time.strftime("%Y-%m-%d_%H-%M-%S")
    time_dir.mkdir(parents=True)

    account = ConnectedDriveAccount(args.username, args.password, get_region_from_name(args.region),
                                    log_responses=time_dir)

    if args.lat and args.lng:
        for vehicle in account.vehicles:
            vehicle.set_observer_position(args.lat, args.lng)
    # doesn't work anymore
    # account.update_vehicle_states()

    # Patching in new My BMW endpoints for fingerprinting
    server_url = get_server_url(get_region_from_name(args.region))

    for vehicle in account.vehicles:
        if vehicle.drive_train in HV_BATTERY_DRIVE_TRAINS:
            print("Getting 'charging-sessions' for {}".format(vehicle.vin))
            account.send_request(
                "https://{}/eadrax-chs/v1/charging-sessions".format(server_url),
                params={
                    "vin": vehicle.vin,
                    "maxResults": 40,
                    "include_date_picker": "true"
                },
                logfilename="charging-sessions"
            )

            print("Getting 'charging-statistics' for {}".format(vehicle.vin))
            account.send_request(
                "https://{}/eadrax-chs/v1/charging-statistics".format(server_url),
                params={
                    "vin": vehicle.vin,
                    "currentDate": datetime.utcnow().isoformat()
                },
                logfilename="charging-statistics"
            )

    print('fingerprint of the vehicles written to {}'.format(time_dir))


def light_flash(args) -> None:
    """Trigger the vehicle to flash its lights."""
    account = ConnectedDriveAccount(args.username, args.password, get_region_from_name(args.region))
    vehicle = account.get_vehicle(args.vin)
    if not vehicle:
        valid_vins = ", ".join(v.vin for v in account.vehicles)
        print('Error: Could not find vehicle for VIN "{}". Valid VINs are: {}'.format(args.vin, valid_vins))
        return
    status = vehicle.remote_services.trigger_remote_light_flash()
    print(status.state)


def vehicle_finder(args) -> None:
    """Trigger the vehicle finder to locate it."""
    account = ConnectedDriveAccount(args.username, args.password, get_region_from_name(args.region))
    account.set_observer_position(args.lat, args.lng)
    vehicle = account.get_vehicle(args.vin)
    if not vehicle:
        valid_vins = ", ".join(v.vin for v in account.vehicles)
        print('Error: Could not find vehicle for VIN "{}". Valid VINs are: {}'.format(args.vin, valid_vins))
        return
    status = vehicle.remote_services.trigger_remote_vehicle_finder()
    print(status.state)
    print({"gps_position": vehicle.status.gps_position, "heading": vehicle.status.gps_heading})


def image(args) -> None:
    """Download a rendered image of the vehicle."""
    account = ConnectedDriveAccount(args.username, args.password, get_region_from_name(args.region))
    vehicle = account.get_vehicle(args.vin)

    for viewdirection in VehicleViewDirection:
        filename = str(viewdirection.name).lower() + '.png'
        with open(filename, 'wb') as output_file:
            image_data = vehicle.get_vehicle_image(viewdirection)
            output_file.write(image_data)
        print('vehicle image saved to {}'.format(filename))


def send_poi(args) -> None:
    """Send Point Of Interest to car."""
    account = ConnectedDriveAccount(args.username, args.password, get_region_from_name(args.region))
    vehicle = account.get_vehicle(args.vin)
    poi_data = dict(
        lat=args.latitude,
        lon=args.longitude,
        name=args.name,
        street=args.street,
        city=args.city,
        postal_code=args.postalcode,
        country=args.country
    )
    vehicle.remote_services.trigger_send_poi(poi_data)


def send_poi_from_address(args) -> None:
    """Create Point of Interest from OSM Nominatim and send to car."""
    account = ConnectedDriveAccount(args.username, args.password, get_region_from_name(args.region))
    vehicle = account.get_vehicle(args.vin)
    address = [(str(' '.join(args.address)))]
    try:
        response = requests.get("https://nominatim.openstreetmap.org",
                                params={
                                    "q": address,
                                    "format": "json",
                                    "addressdetails": 1,
                                    "limit": 1
                                }).json()[0]
    except IndexError:
        print('\nAddress not found')
        sys.exit(1)
    address = response.get("address", {})
    city = address.get("city")
    town = address.get("town")

    poi_data = dict(
        lat=response["lat"],
        lon=response["lon"],
        name=args.name,
        street=address.get("road"),
        city=town if city is None and town is not None else None,
        postal_code=address.get("postcode"),
        country=address.get("country")
    )
    vehicle.remote_services.trigger_send_poi(poi_data)


def send_message(args) -> None:
    """Send a message to car."""
    account = ConnectedDriveAccount(args.username, args.password, get_region_from_name(args.region))
    vehicle = account.get_vehicle(args.vin)
    msg_data = dict(
        text=args.text,
        subject=args.subject
    )
    vehicle.remote_services.trigger_send_message(msg_data)


def _add_default_arguments(parser: argparse.ArgumentParser):
    """Add the default arguments username, password, region to the parser."""
    parser.add_argument('username', help='Connected Drive username')
    parser.add_argument('password', help='Connected Drive password')
    parser.add_argument('region', choices=valid_regions(), help='Region of the Connected Drive account')


def _add_position_arguments(parser: argparse.ArgumentParser):
    """Add the lat and lng attributes to the parser."""
    parser.add_argument('lat', type=float, nargs='?', const=0.0,
                        help='(optional) Your current GPS latitude (as float)')
    parser.add_argument('lng', type=float, nargs='?', const=0.0,
                        help='(optional) Your current GPS longitude (as float)')
    parser.set_defaults(func=get_status)


def main():
    """Main function."""
    parser = main_parser()
    args = parser.parse_args()
    args.func(args)


if __name__ == '__main__':
    main()



"""Models the charging profiles of a vehicle."""

import logging
from typing import TYPE_CHECKING, List
from enum import Enum

from bimmer_connected.utils import SerializableBaseClass

if TYPE_CHECKING:
    from bimmer_connected.vehicle_status import VehicleStatus

_LOGGER = logging.getLogger(__name__)


class ChargingMode(str, Enum):
    """Charging mode of electric vehicle."""
    IMMEDIATE_CHARGING = 'immediateCharging'
    DELAYED_CHARGING = 'delayedCharging'


class ChargingPreferences(str, Enum):
    """Charging preferences of electric vehicle."""
    NO_PRESELECTION = 'noPreSelection'
    CHARGING_WINDOW = 'chargingWindow'


class TimerTypes(str, Enum):
    """Different Timer-Types."""
    TWO_WEEKS = 'twoWeeksTimer'
    ONE_WEEK = 'weeklyPlanner'
    OVERRIDE_TIMER = 'overrideTimer'


class ChargingWindow(SerializableBaseClass):
    """
    This class provides a nicer API than parsing the JSON format directly.
    """

    def __init__(self, window_dict: dict):
        self._window_dict = window_dict

    @property
    def start_time(self) -> str:
        """Start of the charging window."""
        # end of reductionOfChargeCurrent == start of charging window
        return "{}:{}".format(
            str(self._window_dict["end"]["hour"]).zfill(2),
            str(self._window_dict["end"]["minute"]).zfill(2),
        )

    @property
    def end_time(self) -> str:
        """End of the charging window."""
        # start of reductionOfChargeCurrent == end of charging window
        return "{}:{}".format(
            str(self._window_dict["start"]["hour"]).zfill(2),
            str(self._window_dict["start"]["minute"]).zfill(2),
        )


class DepartureTimer(SerializableBaseClass):
    """
    This class provides a nicer API than parsing the JSON format directly.
    """

    def __init__(self, timer_dict: dict):
        self._timer_dict = timer_dict

    @property
    def timer_id(self) -> int:
        """ID of this timer."""
        return self._timer_dict.get("id")

    @property
    def start_time(self) -> str:
        """Deperture time for this timer."""
        if "timeStamp" not in self._timer_dict:
            return None
        return "{}:{}".format(
            str(self._timer_dict["timeStamp"]["hour"]).zfill(2),
            str(self._timer_dict["timeStamp"]["minute"]).zfill(2),
        )

    @property
    def action(self) -> bool:
        """What does the timer do."""
        return self._timer_dict.get("action")

    @property
    def weekdays(self) -> List[str]:
        """Active weekdays for this timer."""
        return self._timer_dict.get("timerWeekDays")


def backend_parameter(func):
    """Decorator for parameters reading data from the backend.

    Errors are handled in a default way.
    """
    def _func_wrapper(self: 'ChargingProfile', *args, **kwargs):
        # pylint: disable=protected-access
        if self.charging_profile is None:
            raise ValueError('No data available for vehicles charging profile!')
        try:
            return func(self, *args, **kwargs)
        except KeyError:
            _LOGGER.debug('No data available for attribute %s!', str(func))
            return None
    return _func_wrapper


class ChargingProfile(SerializableBaseClass):  # pylint: disable=too-many-public-methods
    """Models the charging profile of a vehicle."""

    def __init__(self, status: "VehicleStatus"):
        """Constructor."""
        self.charging_profile = status.status["chargingProfile"]

    def __getattr__(self, item):
        """Generic get function for all backend attributes."""
        return self.charging_profile[item]

    @property
    @backend_parameter
    def is_pre_entry_climatization_enabled(self) -> bool:
        """Get status of pre-entry climatization."""
        return bool(self.charging_profile['climatisationOn'])

    @property
    @backend_parameter
    def timer(self) -> dict:
        """List of timer messages."""
        timer_list = {}
        for timer_dict in self.charging_profile["departureTimes"]:
            curr_timer = DepartureTimer(timer_dict)
            timer_list[curr_timer.timer_id] = curr_timer
        return timer_list

    @property
    @backend_parameter
    def preferred_charging_window(self) -> ChargingWindow:
        """Returns the preferred charging window."""
        return ChargingWindow(self.charging_profile['reductionOfChargeCurrent'])

    @property
    @backend_parameter
    def charging_preferences(self) -> str:
        """Returns the prefered charging preferences."""
        return ChargingPreferences(self.charging_profile['chargingPreference'])

    @property
    @backend_parameter
    def charging_mode(self) -> str:
        """Returns the prefered charging mode."""
        return ChargingMode(self.charging_profile['chargingMode'])



"""General utils and base classes used in the library."""

import base64
import datetime
import hashlib
import inspect
import json
import logging
import random
import string
import sys
from abc import ABC
from collections import OrderedDict

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

_LOGGER = logging.getLogger(__name__)

UNICODE_CHARACTER_SET = string.ascii_letters + string.digits + '-._~'


def generate_token(length=30, chars=UNICODE_CHARACTER_SET):
    """Generate a random token with given length and characters."""
    rand = random.SystemRandom()
    return ''.join(rand.choice(chars) for _ in range(length))


def create_s256_code_challenge(code_verifier):
    """Create S256 code_challenge with the given code_verifier."""
    data = hashlib.sha256(code_verifier.encode('ascii')).digest()
    return base64.urlsafe_b64encode(data).rstrip(b'=').decode('UTF-8')


class RetrySession(requests.Session):
    """A `requests.Session` with `Retry`."""

    def __init__(
        self,
        retries=3,
        backoff_factor=0.3,
        status_forcelist=None,
        allowed_methods=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
            allowed_methods=allowed_methods,
        )
        self.adapters = OrderedDict()
        self.mount('http://', HTTPAdapter(max_retries=retry))
        self.mount('https://', HTTPAdapter(max_retries=retry))


def serialize_for_json(obj: object, excluded: list = None, exclude_hidden: bool = True) -> dict:
    """
    Returns all object attributes and properties as dictionary.

    :param excluded list: attributes and parameters NOT to export
    :param exclude_hidden bool: if true, do not export attributes or parameters starting with '_'
    """
    excluded = excluded if excluded else []
    return dict(
        {
            k: v for k, v in obj.__dict__.items()
            if k not in excluded and ((exclude_hidden and not str(k).startswith("_")) or not exclude_hidden)
        },
        **{a: getattr(obj, a) for a in get_class_property_names(obj) if a not in excluded + ["to_json"]}
    )


def get_class_property_names(obj: object):
    """Returns the names of all properties of a class."""
    return [
        p[0] for p in inspect.getmembers(type(obj), inspect.isdatadescriptor)
        if not p[0].startswith("_")
    ]


def to_json(obj: object, *args, **kwargs):
    """Serialize a nested object to json. Tries to call `to_json` attribute on object first."""
    def serialize(obj: object):
        if hasattr(obj, 'as_dict'):
            return getattr(obj, "as_dict")()
        if hasattr(obj, '__dict__'):
            return {k: v for k, v in getattr(obj, "__dict__").items() if not k.startswith("_")}
        return str(obj)

    return json.dumps(obj, default=serialize, *args, **kwargs)


def parse_datetime(date_str: str) -> datetime.datetime:
    """Convert a time string into datetime."""
    if not date_str:
        return None
    date_formats = ["%Y-%m-%dT%H:%M:%S.%f%z", "%Y-%m-%dT%H:%M:%S%z", "%Y-%m-%dT%H:%M:%S.%fZ", "%Y-%m-%dT%H:%M:%SZ"]
    for date_format in date_formats:
        try:
            parsed = datetime.datetime.strptime(date_str, date_format)
            # Assume implicit UTC for Python 3.6
            if sys.version_info < (3, 7):
                parsed = parsed.replace(tzinfo=datetime.timezone.utc)
            return parsed
        except ValueError:
            pass
    _LOGGER.error("unable to parse '%s' using %s", date_str, date_formats)
    return None


class SerializableBaseClass(ABC):  # pylint: disable=too-few-public-methods
    """Base class to enable json-compatible serialization."""

    def as_dict(self) -> dict:
        """Return all attributes and parameters."""
        return serialize_for_json(self)



"""URLs for different services and error code mapping."""

AUTH_URL = '{gcdm_base_url}/gcdm/oauth/authenticate'
X_USER_AGENT = 'android(v1.07_20200330);{};1.7.0(11152)'

BASE_URL = 'https://{server}'

AUTH_CHINA_PUBLIC_KEY_URL = BASE_URL + '/eadrax-coas/v1/cop/publickey'
AUTH_CHINA_LOGIN_URL = BASE_URL + '/eadrax-coas/v1/login/pwd'

OAUTH_CONFIG_URL = BASE_URL + '/eadrax-ucs/v1/presentation/oauth/config'

VEHICLES_URL = BASE_URL + '/eadrax-vcs/v1/vehicles'

REMOTE_SERVICE_BASE_URL = BASE_URL + '/eadrax-vrccs/v2/presentation/remote-commands'
REMOTE_SERVICE_URL = REMOTE_SERVICE_BASE_URL + '/{vin}/{service_type}'
REMOTE_SERVICE_STATUS_URL = REMOTE_SERVICE_BASE_URL + '/eventStatus?eventId={event_id}'
REMOTE_SERVICE_POSITION_URL = REMOTE_SERVICE_BASE_URL + '/eventPosition?eventId={event_id}'

VEHICLE_IMAGE_URL = BASE_URL + "/eadrax-ics/v3/presentation/vehicles/{vin}/images?carView={view}"
VEHICLE_POI_URL = BASE_URL + '/eadrax-dcs/v1/send-to-car/send-to-car'

VEHICLE_CHARGING_STATISTICS_URL = BASE_URL + '/eadrax-chs/v1/charging-statistics'
VEHICLE_CHARGING_SESSIONS_URL = BASE_URL + '/eadrax-chs/v1/charging-sessions'

SERVICE_PROPERTIES = 'properties'
SERVICE_STATUS = 'status'
SERVICE_CHARGING_STATISTICS_URL = 'CHARGING_STATISTICS'
SERVICE_CHARGING_SESSIONS_URL = 'CHARGING_SESSIONS'
SERVICE_CHARGING_PROFILE = 'CHARGING_PROFILE'

# Possible error codes, other codes are mapped to UNKNOWN_ERROR
ERROR_CODE_MAPPING = {
    401: 'UNAUTHORIZED',
    404: 'NOT_FOUND',
    405: 'MOBILE_ACCESS_DISABLED',
    408: 'VEHICLE_UNAVAILABLE',
    423: 'ACCOUNT_LOCKED',
    429: 'TOO_MANY_REQUESTS',
    500: 'SERVER_ERROR',
    503: 'SERVICE_MAINTENANCE',
}



"""Test for VehicleState."""

import datetime
import logging
import unittest
import time_machine

from bimmer_connected.account import ConnectedDriveAccount
from bimmer_connected.country_selector import get_region_from_name
from bimmer_connected.vehicle_status import (
    LidState,
    LockState,
    ConditionBasedServiceStatus,
    ChargingState,
    VehicleStatus,
)

from . import VIN_F11, VIN_F31, VIN_F48, VIN_G01, VIN_G08, VIN_G30, VIN_I01_REX
from .test_account import get_mocked_account


class TestState(unittest.TestCase):
    """Test for VehicleState."""

    # pylint: disable=protected-access,too-many-public-methods

    def test_generic(self):
        """Test generic attributes."""
        status = get_mocked_account().get_vehicle(VIN_G30).status

        expected = datetime.datetime(
            year=2021, month=11, day=11, hour=8, minute=58, second=53, tzinfo=datetime.timezone.utc
        )
        self.assertEqual(expected, status.timestamp)

        self.assertEqual(7991, status.mileage[0])
        self.assertEqual("km", status.mileage[1])

        self.assertTupleEqual((12.3456, 34.5678), status.gps_position)
        self.assertAlmostEqual(123, status.gps_heading)

    def test_range_combustion_no_info(self):
        """Test if the parsing of mileage and range is working"""
        status = get_mocked_account().get_vehicle(VIN_F31).status

        self.assertTupleEqual((32, "LITERS"), status.remaining_fuel)
        self.assertTupleEqual((None, None), status.remaining_range_fuel)
        self.assertIsNone(status.fuel_percent)

        self.assertIsNone(status.charging_level_hv)
        self.assertTupleEqual((None, None), status.remaining_range_electric)

        self.assertTupleEqual((None, None), status.remaining_range_total)

    def test_range_combustion(self):
        """Test if the parsing of mileage and range is working"""
        status = get_mocked_account().get_vehicle(VIN_F48).status

        self.assertTupleEqual((19, "LITERS"), status.remaining_fuel)
        self.assertTupleEqual((308, "km"), status.remaining_range_fuel)
        self.assertIsNone(status.fuel_percent)

        self.assertIsNone(status.charging_level_hv)
        self.assertTupleEqual((None, None), status.remaining_range_electric)

        self.assertTupleEqual((308, "km"), status.remaining_range_total)

    def test_range_phev(self):
        """Test if the parsing of mileage and range is working"""
        status = get_mocked_account().get_vehicle(VIN_G30).status

        self.assertTupleEqual((11, "LITERS"), status.remaining_fuel)
        self.assertTupleEqual((107, "km"), status.remaining_range_fuel)
        self.assertAlmostEqual(28, status.fuel_percent)

        self.assertAlmostEqual(41, status.charging_level_hv)
        self.assertTupleEqual((9, "km"), status.remaining_range_electric)

        self.assertTupleEqual((116, "km"), status.remaining_range_total)

        self.assertAlmostEqual(
            status.remaining_range_fuel[0] + status.remaining_range_electric[0], status.remaining_range_total[0]
        )

    def test_range_rex(self):
        """Test if the parsing of mileage and range is working"""
        status = get_mocked_account().get_vehicle(VIN_I01_REX).status

        self.assertTupleEqual((5, "LITERS"), status.remaining_fuel)
        self.assertTupleEqual((64, "km"), status.remaining_range_fuel)
        self.assertIsNone(status.fuel_percent)

        self.assertAlmostEqual(100, status.charging_level_hv)
        self.assertTupleEqual((164, "km"), status.remaining_range_electric)

        self.assertTupleEqual((228, "km"), status.remaining_range_total)

        self.assertAlmostEqual(
            status.remaining_range_fuel[0] + status.remaining_range_electric[0], status.remaining_range_total[0]
        )

    def test_range_electric(self):
        """Test if the parsing of mileage and range is working"""
        status = get_mocked_account().get_vehicle(VIN_G08).status

        self.assertTupleEqual((0, "LITERS"), status.remaining_fuel)
        self.assertTupleEqual((None, None), status.remaining_range_fuel)
        self.assertEqual(0, status.fuel_percent)

        self.assertAlmostEqual(50, status.charging_level_hv)
        self.assertTupleEqual((179, "km"), status.remaining_range_electric)

        self.assertTupleEqual((179, "km"), status.remaining_range_total)

    @time_machine.travel(
        datetime.datetime.now().replace(
            hour=21, minute=28, second=59, microsecond=0, tzinfo=ConnectedDriveAccount.timezone()
        )
    )
    def test_remaining_charging_time(self):
        """Test if the parsing of mileage and range is working"""
        account = get_mocked_account()
        status = account.get_vehicle(VIN_G08).status
        self.assertEqual(6.53, status.charging_time_remaining)

    @time_machine.travel("2011-11-28 21:28:59 +0000", tick=False)
    def test_charging_end_time(self):
        """Test if the parsing of mileage and range is working"""
        account = get_mocked_account()
        status = account.get_vehicle(VIN_G08).status
        self.assertEqual(
            datetime.datetime(2011, 11, 29, 4, 1, tzinfo=ConnectedDriveAccount.timezone()), status.charging_end_time
        )

    def test_charging_time_label(self):
        """Test if the parsing of mileage and range is working"""
        account = get_mocked_account()
        status = account.get_vehicle(VIN_G08).status
        self.assertEqual("100% at ~04:01 AM", status.charging_time_label)

    def test_charging_end_time_parsing_failure(self):
        """Test if the parsing of mileage and range is working"""
        account = get_mocked_account()
        vehicle = account.get_vehicle(VIN_G08)
        with self.assertLogs(level=logging.ERROR):
            vehicle.update_state(
                {
                    "status": {
                        "fuelIndicators": [
                            {
                                "chargingStatusIndicatorType": "CHARGING",
                                "chargingStatusType": "CHARGING",
                                "infoLabel": "100% at later today...",
                                "rangeIconId": 59683,
                                "rangeUnits": "km",
                                "rangeValue": "179",
                            }
                        ]
                    },
                    "properties": {},
                }
            )
        self.assertIsNone(vehicle.status.charging_end_time)
        self.assertEqual("100% at later today...", vehicle.status.charging_time_label)
        self.assertEqual(0, vehicle.status.charging_time_remaining)

    def test_plugged_in_waiting_for_charge_window(self):
        """G01 is plugged in but not charging, as its waiting for charging window."""
        # Should be None on G01 as it is only "charging"
        account = get_mocked_account()
        vehicle = account.get_vehicle(VIN_G01)

        self.assertIsNone(vehicle.status.charging_end_time)
        self.assertEqual("Starts at ~ 09:00 AM", vehicle.status.charging_time_label)
        self.assertEqual(0, vehicle.status.charging_time_remaining)
        self.assertEqual(ChargingState.PLUGGED_IN, vehicle.status.charging_status)
        self.assertEqual("CONNECTED", vehicle.status.connection_status)

    def test_condition_based_services(self):
        """Test condition based service messages."""
        status = get_mocked_account().get_vehicle(VIN_G30).status

        cbs = status.condition_based_services
        self.assertEqual(3, len(cbs))
        self.assertEqual(ConditionBasedServiceStatus.OK, cbs[0].state)
        expected_cbs0 = datetime.datetime(year=2022, month=8, day=1, tzinfo=datetime.timezone.utc)
        self.assertEqual(expected_cbs0, cbs[0].due_date)
        self.assertTupleEqual((25000, "KILOMETERS"), cbs[0].due_distance)

        self.assertEqual(ConditionBasedServiceStatus.OK, cbs[1].state)
        expected_cbs1 = datetime.datetime(year=2023, month=8, day=1, tzinfo=datetime.timezone.utc)
        self.assertEqual(expected_cbs1, cbs[1].due_date)
        self.assertIsNone(cbs[1].due_distance)

        self.assertEqual(ConditionBasedServiceStatus.OK, cbs[2].state)
        expected_cbs2 = datetime.datetime(year=2024, month=8, day=1, tzinfo=datetime.timezone.utc)
        self.assertEqual(expected_cbs2, cbs[2].due_date)
        self.assertTupleEqual((60000, "KILOMETERS"), cbs[2].due_distance)

        self.assertTrue(status.are_all_cbs_ok)

    def test_parse_f31_no_position(self):
        """Test parsing of F31 data with position tracking disabled in the vehicle."""
        status = get_mocked_account().get_vehicle(VIN_F31).status

        self.assertTupleEqual((None, None), status.gps_position)
        self.assertIsNone(status.gps_heading)

    def test_parse_gcj02_position(self):
        """Test conversion of GCJ02 to WGS84 for china."""
        account = get_mocked_account(get_region_from_name("china"))
        status = VehicleStatus(
            account,
            {
                "properties": {
                    "vehicleLocation": {
                        "address": {"formatted": "some_formatted_address"},
                        "coordinates": {"latitude": 39.83492, "longitude": 116.23221},
                        "heading": 123,
                    },
                    "lastUpdatedAt": "2021-11-14T20:20:21Z",
                },
                "status": {
                    "fuelIndicators": [],
                    "lastUpdatedAt": "2021-11-14T20:20:21Z",
                },
            },
        )
        self.assertTupleEqual(
            (39.8337, 116.22617), (round(status.gps_position[0], 5), round(status.gps_position[1], 5))
        )

    def test_parse_f11_no_position_vehicle_active(self):
        """Test parsing of F11 data with vehicle beeing active."""
        vehicle = get_mocked_account().get_vehicle(VIN_F48)
        status = vehicle.status

        self.assertTrue(vehicle.is_vehicle_tracking_enabled)
        self.assertTrue(status.is_vehicle_active)
        with self.assertLogs(level=logging.INFO):
            self.assertTupleEqual((None, None), status.gps_position)
            self.assertIsNone(status.gps_heading)

    def test_parse_g08(self):
        """Test if the parsing of the attributes is working."""
        status = get_mocked_account().get_vehicle(VIN_G08).status

        self.assertTupleEqual((179, "km"), status.remaining_range_electric)
        self.assertTupleEqual((179, "km"), status.remaining_range_total)
        self.assertEqual(ChargingState.CHARGING, status.charging_status)
        self.assertEqual(50, status.charging_level_hv)

    # def test_missing_attribute(self):
    #     """Test if error handling is working correctly."""
    #     account = unittest.mock.MagicMock(ConnectedDriveAccount)
    #     state = VehicleState(account, None)
    #     state._attributes[SERVICE_STATUS] = {}
    #     self.assertIsNone(status.mileage)

    def test_lids(self):
        """Test features around lids."""
        status = get_mocked_account().get_vehicle(VIN_G30).status

        self.assertEqual(6, len(list(status.lids)))
        self.assertEqual(3, len(list(status.open_lids)))
        self.assertFalse(status.all_lids_closed)

        status = get_mocked_account().get_vehicle(VIN_G08).status

        for lid in status.lids:
            self.assertEqual(LidState.CLOSED, lid.state)
        self.assertTrue(status.all_lids_closed)
        self.assertEqual(6, len(list(status.lids)))

    def test_windows_g31(self):
        """Test features around windows."""
        status = get_mocked_account().get_vehicle(VIN_G08).status

        for window in status.windows:
            self.assertEqual(LidState.CLOSED, window.state)

        self.assertEqual(5, len(list(status.windows)))
        self.assertEqual(0, len(list(status.open_windows)))
        self.assertTrue(status.all_windows_closed)

    def test_door_locks(self):
        """Test the door locks."""
        status = get_mocked_account().get_vehicle(VIN_G08).status

        self.assertEqual(LockState.LOCKED, status.door_lock_state)

        status = get_mocked_account().get_vehicle(VIN_I01_REX).status

        self.assertEqual(LockState.UNLOCKED, status.door_lock_state)

    def test_empty_status(self):
        """Test an empy status."""
        status = get_mocked_account().get_vehicle(VIN_F31).status
        status.properties.pop("inMotion")

        with self.assertLogs(level=logging.DEBUG):
            self.assertIsNone(status.is_vehicle_active)

        status.properties = None
        status.status = None

        with self.assertRaises(ValueError):
            self.assertIsNone(status.is_vehicle_active)

    # def test_parsing_attributes(self):
    #     """Test parsing different attributes of the vehicle.

    #     Just make sure parsing that no exception is raised and we get not-None values.
    #     """
    #     backend_mock = BackendMock()
    #     # list of attributes that are ignored at the moment
    #     ignored_attributes = [ATTRIBUTE_MAPPING.get(a, a) for a in MISSING_ATTRIBUTES]
    #     with mock.patch('bimmer_connected.account.requests', new=backend_mock):
    #         backend_mock.setup_default_vehicles()
    #         account = ConnectedDriveAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)
    #         account.update_vehicle_states()

    #         for vehicle in account.vehicles:
    #             print(vehicle.name)
    #             for attribute in (a for a in vehicle.available_attributes if a not in ignored_attributes):
    #                 self.assertIsNotNone(getattr(vehicle.status, attribute), attribute)

    def test_check_control_messages(self):
        """Test handling of check control messages.

        G21 is the only vehicle with active Check Control Messages, so we only expect to get something there.
        However we have no vehicle with issues in check control.
        """
        vehicle = get_mocked_account().get_vehicle(VIN_F11)
        self.assertTrue(vehicle.status.has_check_control_messages)

        ccms = vehicle.status.check_control_messages
        self.assertEqual(2, len(ccms))

        self.assertEqual("Medium", ccms[0].state)
        self.assertEqual("229", ccms[0].ccm_id)
        self.assertEqual(
            (
                "Charge by driving for longer periods or use external charger. "
                "Functions requiring battery will be switched off."
            ),
            ccms[0].description_long,
        )
        self.assertEqual("Battery discharged: Start engine", ccms[0].description_short)

        self.assertEqual("Low", ccms[1].state)
        self.assertEqual("50", ccms[1].ccm_id)
        self.assertEqual(
            (
                "System unable to monitor tire pressure. Check tire pressures manually. "
                "Continued driving possible. Consult service center."
            ),
            ccms[1].description_long,
        )
        self.assertEqual("Flat Tire Monitor (FTM) inactive", ccms[1].description_short)



"""Tests for ConnectedDriveAccount."""
import json
import unittest

import requests_mock
from requests import HTTPError

from bimmer_connected.account import ConnectedDriveAccount
from bimmer_connected.country_selector import get_region_from_name

from . import (
    RESPONSE_DIR,
    TEST_PASSWORD,
    TEST_REGION,
    TEST_REGION_STRING,
    TEST_USERNAME,
    VIN_G21,
    get_fingerprint_count,
    load_response,
)


def authenticate_callback(request, context):  # pylint: disable=inconsistent-return-statements
    """Returns /oauth/authentication response based on request."""
    # pylint: disable=protected-access,unused-argument,no-self-use

    if "username" in request.text and "password" in request.text and "grant_type" in request.text:
        return load_response(RESPONSE_DIR / "auth" / "authorization_response.json")
    context.headers = {
        "Location": "com.mini.connected://oauth?code=CODE&state=STATE&client_id=CLIENT_ID&nonce=login_nonce",
    }
    context.status_code = 302


def return_vehicles(request, context):  # pylint: disable=inconsistent-return-statements
    """Returns /vehicles response based on x-user-agent."""
    # pylint: disable=protected-access,unused-argument,no-self-use

    x_user_agent = request._request.headers.get("x-user-agent", "").split(";")
    if len(x_user_agent) == 3:
        brand = x_user_agent[1]
    else:
        raise ValueError("x-user-agent not configured correctly!")

    response_vehicles = []
    files = RESPONSE_DIR.rglob("vehicles_v2_{}_0.json".format(brand))
    for file in files:
        response_vehicles.extend(load_response(file))
    return response_vehicles


def get_base_adapter():
    """Returns mocked adapter for auth."""
    adapter = requests_mock.Adapter()
    adapter.register_uri(
        "GET",
        "/eadrax-ucs/v1/presentation/oauth/config",
        json=load_response(RESPONSE_DIR / "auth" / "oauth_config.json"),
    )
    adapter.register_uri("POST", "/gcdm/oauth/authenticate", json=authenticate_callback)
    adapter.register_uri("POST", "/gcdm/oauth/token", json=load_response(RESPONSE_DIR / "auth" / "auth_token.json"))
    adapter.register_uri("GET", "/eadrax-vcs/v1/vehicles", json=return_vehicles)
    adapter.register_uri(
        "GET", "/eadrax-coas/v1/cop/publickey", json=load_response(RESPONSE_DIR / "auth" / "auth_cn_publickey.json")
    )
    adapter.register_uri(
        "POST", "/eadrax-coas/v1/login/pwd", json=load_response(RESPONSE_DIR / "auth" / "auth_cn_login_pwd.json")
    )
    return adapter


def get_mocked_account(region=None):
    """Returns pre-mocked account."""
    with requests_mock.Mocker(adapter=get_base_adapter()):
        account = ConnectedDriveAccount(TEST_USERNAME, TEST_PASSWORD, region or TEST_REGION)
    return account


class TestAccount(unittest.TestCase):
    """Tests for ConnectedDriveAccount."""

    def test_login_row_na(self):
        """Test the login flow."""
        with requests_mock.Mocker(adapter=get_base_adapter()):
            account = ConnectedDriveAccount(TEST_USERNAME, TEST_PASSWORD, get_region_from_name(TEST_REGION_STRING))
        self.assertIsNotNone(account)

    def test_login_china(self):
        """Test raising an error for region `china`."""
        with requests_mock.Mocker(adapter=get_base_adapter()):
            account = ConnectedDriveAccount(TEST_USERNAME, TEST_PASSWORD, get_region_from_name("china"))
        self.assertIsNotNone(account)

    def test_vehicles(self):
        """Test the login flow."""
        account = get_mocked_account()

        self.assertIsNotNone(account._oauth_token)  # pylint: disable=protected-access
        self.assertEqual(get_fingerprint_count(), len(account.vehicles))
        vehicle = account.get_vehicle(VIN_G21)
        self.assertEqual(VIN_G21, vehicle.vin)

        self.assertIsNone(account.get_vehicle("invalid_vin"))

    def test_invalid_password(self):
        """Test parsing the results of an invalid password."""
        with requests_mock.Mocker(adapter=get_base_adapter()) as mock:
            mock.post(
                "/gcdm/oauth/authenticate",
                json=load_response(RESPONSE_DIR / "auth" / "auth_error_wrong_password.json"),
                status_code=401,
            )
            with self.assertRaises(HTTPError):
                ConnectedDriveAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)

    def test_invalid_password_china(self):
        """Test parsing the results of an invalid password."""
        with requests_mock.Mocker(adapter=get_base_adapter()) as mock:
            mock.post(
                "/eadrax-coas/v1/login/pwd",
                json=load_response(RESPONSE_DIR / "auth" / "auth_cn_login_error.json"),
                status_code=422,
            )
            with self.assertRaises(HTTPError):
                ConnectedDriveAccount(TEST_USERNAME, TEST_PASSWORD, get_region_from_name("china"))

    def test_server_error(self):
        """Test parsing the results of a server error."""
        with requests_mock.Mocker(adapter=get_base_adapter()) as mock:
            mock.post(
                "/gcdm/oauth/authenticate",
                text=load_response(RESPONSE_DIR / "auth" / "auth_error_internal_error.txt"),
                status_code=500,
            )
            with self.assertRaises(HTTPError):
                ConnectedDriveAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)

    def test_anonymize_data(self):
        """Test anonymization function."""
        test_dict = {
            "vin": "secret",
            "a sub-dict": {
                "lat": 666,
                "lon": 666,
                "heading": 666,
            },
            "licensePlate": "secret",
            "public": "public_data",
            "a_list": [
                {"vin": "secret"},
                {
                    "lon": 666,
                    "public": "more_public_data",
                },
            ],
            "b_list": ["a", "b"],
            "empty_list": [],
        }
        anon_text = json.dumps(ConnectedDriveAccount._anonymize_data(test_dict))  # pylint: disable=protected-access
        self.assertNotIn("secret", anon_text)
        self.assertNotIn("666", anon_text)
        self.assertIn("public_data", anon_text)
        self.assertIn("more_public_data", anon_text)

    def test_vehicle_search_case(self):
        """Check if the search for the vehicle by VIN is NOT case sensitive."""
        account = get_mocked_account()

        vin = account.vehicles[1].vin
        self.assertEqual(vin, account.get_vehicle(vin).vin)
        self.assertEqual(vin, account.get_vehicle(vin.lower()).vin)
        self.assertEqual(vin, account.get_vehicle(vin.upper()).vin)

    def test_set_observer_value(self):
        """Test set_observer_position with valid arguments."""
        account = get_mocked_account()

        account.set_observer_position(1.0, 2.0)
        for vehicle in account.vehicles:
            self.assertEqual(vehicle.observer_latitude, 1.0)
            self.assertEqual(vehicle.observer_longitude, 2.0)

    def test_set_observer_not_set(self):
        """Test set_observer_position with no arguments."""
        account = get_mocked_account()

        for vehicle in account.vehicles:
            self.assertIsNone(vehicle.observer_latitude)
            self.assertIsNone(vehicle.observer_longitude)

        account.set_observer_position(17.99, 179.9)

        for vehicle in account.vehicles:
            self.assertEqual(vehicle.observer_latitude, 17.99)
            self.assertEqual(vehicle.observer_longitude, 179.9)

    def test_set_observer_some_none(self):
        """Test set_observer_position with invalid arguments."""
        account = get_mocked_account()

        with self.assertRaises(ValueError):
            account.set_observer_position(None, 2.0)

        with self.assertRaises(ValueError):
            account.set_observer_position(1.0, None)



"""Library to read data from the BMW Connected Drive portal.

The library bimmer_connected provides a Python interface to interact
with the BMW Connected Drive web service. It allows you to read
the current state of the vehicle and also trigger remote services.

Disclaimer:
This library is not affiliated with or endorsed by BMW Group.
"""

import base64
import datetime
import json
import logging
import pathlib
import urllib
from threading import Lock
from typing import Any, Callable, Dict, List

import jwt
import requests
from Crypto.Cipher import PKCS1_v1_5
from Crypto.PublicKey import RSA
from requests.auth import HTTPBasicAuth
from requests.exceptions import HTTPError
from requests.models import Response

from bimmer_connected.const import (
    AUTH_CHINA_LOGIN_URL,
    AUTH_CHINA_PUBLIC_KEY_URL,
    AUTH_URL,
    OAUTH_CONFIG_URL,
    VEHICLES_URL,
    X_USER_AGENT
)
from bimmer_connected.country_selector import (
    Regions,
    get_ocp_apim_key,
    get_server_url
)
from bimmer_connected.utils import (
    RetrySession,
    create_s256_code_challenge,
    generate_token
)
from bimmer_connected.vehicle import CarBrand, ConnectedDriveVehicle

VALID_UNTIL_OFFSET = datetime.timedelta(seconds=10)

_LOGGER = logging.getLogger(__name__)


class ConnectedDriveAccount:  # pylint: disable=too-many-instance-attributes
    """Create a new connection to the BMW Connected Drive web service.

    :param username: Connected drive user name
    :param password: Connected drive password
    :param country: Country for which the account was created. For a list of valid countries,
                check https://www.bmw-connecteddrive.com .
                Use the name of the countries exactly as on the website.
    :param log_responses: If log_responses is set, all responses from the server will
                be loged into this directory. This can be used for later analysis of the different
                responses for different vehicles.
    :param retries_on_500_error: If retries_on_500_error is set, a communication with the
                Connected Drive server will automatically be retried the number of times
                specified in the event the error code received was 500. This sometimes
                occurs (presumably) due to bugs in the server implementation.
    """

    # pylint: disable=too-many-arguments,too-many-locals
    def __init__(self, username: str, password: str, region: Regions, log_responses: pathlib.Path = None,
                 retries_on_500_error: int = 5) -> None:
        self._region = region
        self._server_url_legacy = None
        self._server_url = None
        self._username = username
        self._password = password
        self._oauth_token = None
        self._refresh_token = None
        self._token_expiration = None
        self._log_responses = log_responses
        self._retries_on_500_error = retries_on_500_error
        #: list of vehicles associated with this account.
        self._vehicles = []  # type: list[ConnectedDriveVehicle]
        self._lock = Lock()
        self._update_listeners = []  # type: list[Callable[[], Any]]

        self._get_vehicles()

    def _get_oauth_token(self) -> None:
        """Get a new auth token from the server."""
        with self._lock:
            if self._token_expiration is not None and datetime.datetime.utcnow() < self._token_expiration:
                _LOGGER.debug('Old token is still valid. Not getting a new one.')
                return

            try:
                if self.region in [Regions.REST_OF_WORLD, Regions.NORTH_AMERICA]:
                    token_data = self._login_row_na()
                if self.region == Regions.CHINA:
                    token_data = self._login_china()

                self._oauth_token = token_data["access_token"]
                self._token_expiration = token_data["expires_at"]
                _LOGGER.debug(
                    "got new token %s with expiration date %s",
                    self._oauth_token,
                    self._token_expiration,
                )

            except Exception as ex:  # pylint: disable=broad-except
                _LOGGER.exception(ex)
                raise ex

    def _login_row_na(self):
        """Login to Rest of World and North America."""
        try:
            # We need a session for cross-request cookies
            oauth_session = RetrySession(status_forcelist=[401], allowed_methods=frozenset(["GET", "POST"]))
            r_oauth_settings = oauth_session.get(
                OAUTH_CONFIG_URL.format(server=self.server_url),
                headers={
                    "ocp-apim-subscription-key": get_ocp_apim_key(self._region),
                    'user-agent': "Dart/2.13 (dart:io)",
                    "x-user-agent": X_USER_AGENT.format("bmw"),
                }
            )
            r_oauth_settings.raise_for_status()
            oauth_settings = r_oauth_settings.json()

            # My BMW login flow
            _LOGGER.debug("Authenticating against GCDM with MyBMW flow.")

            code_verifier = generate_token(86)
            code_challenge = create_s256_code_challenge(code_verifier)

            state = generate_token(22)

            authenticate_url = AUTH_URL.format(gcdm_base_url=oauth_settings["gcdmBaseUrl"])
            authenticate_headers = {
                "Content-Type": "application/x-www-form-urlencoded",
                "Referer": "https://customer.bmwgroup.com/oneid/",
                "User-Agent": (
                    "Mozilla/5.0 (Linux; Android 7.1.2; One) AppleWebKit/537.36 (KHTML, like Gecko) "
                    "Chrome/95.0.4638.74 Mobile Safari/537.36"
                ),
            }

            # we really need all of these parameters
            oauth_base_values = {
                "client_id": oauth_settings["clientId"],
                "response_type": "code",
                "redirect_uri": oauth_settings["returnUrl"],
                "state": state,
                "nonce": "login_nonce",
                "scope": " ".join(oauth_settings["scopes"]),
                "code_challenge": code_challenge,
                "code_challenge_method": "S256",
            }

            authenticate_data = dict(
                oauth_base_values,
                **{
                    "grant_type": "authorization_code",
                    "username": self._username,
                    "password": self._password,
                }
            )
            response = oauth_session.post(
                authenticate_url,
                headers=authenticate_headers,
                data=authenticate_data,
            )
            response.raise_for_status()
            authorization = dict(urllib.parse.parse_qsl(response.json()["redirect_to"]))["authorization"]

            code_data = urllib.parse.urlencode(
                dict(oauth_base_values, **{"authorization": authorization})
            )
            response = oauth_session.post(
                authenticate_url, headers=authenticate_headers, data=code_data, allow_redirects=False
            )
            response.raise_for_status()
            code = dict(urllib.parse.parse_qsl(response.next.path_url.split('?')[1]))["code"]

            token_url = oauth_settings["tokenEndpoint"]

            # My BMW login flow
            token_headers = {
                "Content-Type": "application/x-www-form-urlencoded",
                # "Authorization": oauth_settings["token"]["Authorization"],
            }
            token_values = {
                "code": code,
                "code_verifier": code_verifier,
                "redirect_uri": oauth_settings["returnUrl"],
                "grant_type": "authorization_code",
            }

            token_data = urllib.parse.urlencode(token_values)
            current_utc_time = datetime.datetime.utcnow()
            response = oauth_session.post(
                token_url,
                headers=token_headers,
                data=token_data,
                auth=HTTPBasicAuth(oauth_settings["clientId"], oauth_settings["clientSecret"])
            )
            response.raise_for_status()
            response_json = response.json()

            expiration_time = int(response_json["expires_in"])
            expires_at = current_utc_time + datetime.timedelta(seconds=expiration_time) - VALID_UNTIL_OFFSET

            return {
                "access_token": response_json["access_token"],
                "expires_at": expires_at
            }
        except HTTPError as ex:
            try:
                err = response.json()
                _LOGGER.error("Authentication failed (%s): %s", err["error"], err["error_description"])
            except Exception:  # pylint: disable=broad-except
                _LOGGER.error("Authentication failed: %s", response.text)
            raise ex

    def _login_china(self):
        try:
            login_header = {
                'user-agent': "Dart/2.13 (dart:io)",
                'x-user-agent': X_USER_AGENT.format("bmw")
            }

            response = requests.request(
                "GET",
                AUTH_CHINA_PUBLIC_KEY_URL.format(server=self.server_url),
                headers=login_header,
            )
            response.raise_for_status()
            pem_public_key = response.json()["data"]["value"]

            public_key = RSA.import_key(pem_public_key)
            cipher_rsa = PKCS1_v1_5.new(public_key)
            encrypted = cipher_rsa.encrypt(self._password.encode())
            pw_encrypted = base64.b64encode(encrypted).decode('UTF-8')

            response = requests.request(
                "POST",
                AUTH_CHINA_LOGIN_URL.format(server=self.server_url),
                headers=login_header,
                json={"mobile": self._username, "password": pw_encrypted}
            )
            response.raise_for_status()
            response_json = response.json()["data"]

            decoded_token = jwt.decode(
                response_json["access_token"],
                algorithms=["HS256"],
                options={"verify_signature": False}
            )

            return {
                "access_token": response_json["access_token"],
                "expires_at": datetime.datetime.utcfromtimestamp(decoded_token["exp"]) - VALID_UNTIL_OFFSET
            }
        except HTTPError as ex:
            try:
                err = response.json()
                _LOGGER.error("Authentication failed (%s): %s", err["error"], err["description"])
            except Exception:  # pylint: disable=broad-except
                _LOGGER.error("Authentication failed: %s", response.text)
            raise ex

    def request_header(self, brand: CarBrand = None) -> Dict[str, str]:
        """Generate a header for HTTP requests to the server."""
        brand = brand or CarBrand.BMW
        headers = {
            "accept": "application/json",
            "user-agent": "Dart/2.13 (dart:io)",
            "x-user-agent": X_USER_AGENT.format(brand.value),
            "Authorization": "Bearer {}".format(self._oauth_token),
            "accept-language": "en",
        }
        return headers

    def send_request(self, url: str, data=None, headers=None, post=False, allow_redirects=True,
                     logfilename: str = None, params: dict = None, brand: CarBrand = None) -> Response:
        """Send an http request to the server.

        If the http headers are not set, default headers are generated.
        You can choose if you want a GET or POST request.
        """
        _LOGGER.debug("Request to: %s", url)
        self._get_oauth_token()
        request_headers = self.request_header(brand)
        if headers:
            request_headers.update(headers)

        if post:
            response = requests.post(url, headers=request_headers, data=data, allow_redirects=allow_redirects,
                                     params=params)
        else:
            response = requests.get(url, headers=request_headers, data=data, allow_redirects=allow_redirects,
                                    params=params)

        try:
            response.raise_for_status()
        except HTTPError as ex:
            try:
                err = response.json()
                _LOGGER.error("MyBMW API error '%s' (%s): %s", ex, err["error"], err["description"])
            except Exception:  # pylint: disable=broad-except
                _LOGGER.error("MyBMW API error '%s': %s", ex, response.text)
            raise ex

        self._log_response_to_file(response, logfilename)
        return response

    def _log_response_to_file(self, response: requests.Response, logfilename: str = None) -> None:
        """If a log path is set, log all resonses to a file."""
        if self._log_responses is None or logfilename is None:
            return

        anonymized_data = json.dumps(self._anonymize_data(response.json()), indent=2, sort_keys=True)

        output_path = None
        count = 0

        while output_path is None or output_path.exists():
            output_path = self._log_responses / '{}_{}.txt'.format(logfilename, count)
            count += 1

        with open(output_path, 'w', encoding='UTF-8') as logfile:
            logfile.write(anonymized_data)

    @staticmethod
    def _anonymize_data(json_data: dict) -> dict:
        """Replace parts of the logfiles containing personal information."""

        replacements = {
            'lat': 12.3456,
            'latitude': 12.3456,
            'lon': 34.5678,
            'longitude': 34.5678,
            'heading': 123,
            'vin': 'some_vin',
            'licensePlate': 'some_license_plate',
            'name': 'some_name',
            'city': 'some_city',
            'street': 'some_street',
            'streetNumber': '999',
            'postalCode': 'some_postal_code',
            'phone': 'some_phone',
            'formatted': 'some_formatted_address',
            'subtitle': 'some_road \u2022 duration \u2022 -- EUR',
        }

        if isinstance(json_data, list):
            json_data = [ConnectedDriveAccount._anonymize_data(v) for v in json_data]
        elif isinstance(json_data, dict):
            for key, value in json_data.items():
                if key in replacements:
                    json_data[key] = replacements[key]
                else:
                    json_data[key] = ConnectedDriveAccount._anonymize_data(value)

        return json_data

    @property
    def server_url(self) -> str:
        """Get the url of the server for this country."""
        if self._server_url is None:
            self._server_url = get_server_url(self._region)
        return self._server_url

    @property
    def region(self) -> str:
        """Get the region."""
        return self._region

    def _get_vehicles(self) -> None:
        """Retrieve list of vehicle for the account."""
        _LOGGER.debug('Getting vehicle list')
        self._get_oauth_token()

        for brand in CarBrand:
            response = self.send_request(
                VEHICLES_URL.format(server=self.server_url),
                headers=self.request_header(brand),
                params={
                    "apptimezone": self.utcdiff,
                    "appDateTime": int(datetime.datetime.now().timestamp()*1000),
                    "tireGuardMode": "ENABLED"},
                logfilename="vehicles_v2_{}".format(brand.value),
            )

            for vehicle_dict in response.json():
                # If vehicle already exists, just update it's state
                existing_vehicle = self.get_vehicle(vehicle_dict["vin"])
                if existing_vehicle:
                    existing_vehicle.update_state(vehicle_dict)
                else:
                    self._vehicles.append(ConnectedDriveVehicle(self, vehicle_dict))

    def get_vehicle(self, vin: str) -> ConnectedDriveVehicle:
        """Get vehicle with given VIN.

        The search is NOT case sensitive.
        :param vin: VIN of the vehicle you want to get.
        :return: Returns None if no such vehicle is found.
        """
        for car in self.vehicles:
            if car.vin.upper() == vin.upper():
                return car
        return None

    def update_vehicle_states(self) -> None:
        """Update the state of all vehicles.

        Notify all listeners of the vehicle state update.
        """
        # With MyBMW, we only have to get the vehicles list.
        self._get_vehicles()
        for listener in self._update_listeners:
            listener()

    def add_update_listener(self, listener: Callable[[], Any]) -> None:
        """Add a listener for state updates."""
        self._update_listeners.append(listener)

    @property
    def vehicles(self) -> List[ConnectedDriveVehicle]:
        """Get list of vehicle of this account"""
        return self._vehicles

    def __str__(self) -> str:
        """Use the user name as id for the account class."""
        return '{}: {}'.format(self.__class__, self._username)

    def set_observer_position(self, latitude: float, longitude: float) -> None:
        """Set the position of the observer for all vehicles.

        see VehicleViewDirection.set_observer_position() for more details.
        """
        if latitude is None or longitude is None:
            raise ValueError('Either latitude and longitude are both not None or both are None.')
        for vehicle in self._vehicles:
            vehicle.set_observer_position(latitude, longitude)

    @staticmethod
    def timezone():
        """Returns the current tzinfo."""
        return datetime.datetime.now().astimezone().tzinfo

    @property
    def utcdiff(self):
        """Returns the difference to UTC in minutes."""
        return round(self.timezone().utcoffset(datetime.datetime.now()).seconds / 60, 0)



"""Python package description."""
from setuptools import setup

setup(
    setup_requires=['pbr'],
    pbr=True,
    install_requires=[
        "requests>=2.24.0",
        "urllib3>1.26.5,<2",
        "pycryptodome>=3.4",
        "pyjwt>=2.1.0",
    ]
)



"""Trigger remote services on a vehicle."""

import datetime
import json
from json.decoder import JSONDecodeError
import logging
import time
from typing import TYPE_CHECKING, Dict
from enum import Enum

from bimmer_connected.const import (REMOTE_SERVICE_POSITION_URL,
                                    REMOTE_SERVICE_STATUS_URL,
                                    REMOTE_SERVICE_URL,
                                    VEHICLE_POI_URL)

if TYPE_CHECKING:
    from bimmer_connected.account import ConnectedDriveAccount
    from bimmer_connected.vehicle import ConnectedDriveVehicle

TIME_FORMAT = '%Y-%m-%dT%H:%M:%S.%f'

_LOGGER = logging.getLogger(__name__)

#: time in seconds between polling updates on the status of a remote service
_POLLING_CYCLE = 3

#: maximum number of seconds to wait for the server to return a positive answer
_POLLING_TIMEOUT = 240

#: time in seconds to wait before updating the vehicle state from the server
_UPDATE_AFTER_REMOTE_SERVICE_DELAY = 10


class ExecutionState(str, Enum):
    """Enumeration of possible states of the execution of a remote service."""
    INITIATED = 'INITIATED'
    PENDING = 'PENDING'
    DELIVERED = 'DELIVERED'
    EXECUTED = 'EXECUTED'
    ERROR = 'ERROR'
    UNKNOWN = 'UNKNOWN'


class _Services(str, Enum):
    """Enumeration of possible services to be executed."""
    REMOTE_LIGHT_FLASH = 'LIGHT_FLASH'
    REMOTE_VEHICLE_FINDER = 'VEHICLE_FINDER'
    REMOTE_DOOR_LOCK = 'DOOR_LOCK'
    REMOTE_DOOR_UNLOCK = 'DOOR_UNLOCK'
    REMOTE_HORN = 'HORN_BLOW'
    REMOTE_AIR_CONDITIONING = 'CLIMATE_NOW'
    REMOTE_CHARGE_NOW = 'CHARGE_NOW'


class RemoteServiceStatus:  # pylint: disable=too-few-public-methods
    """Wraps the status of the execution of a remote service."""

    def __init__(self, response: dict):
        """Construct a new object from a dict."""
        status = None
        if 'executionStatus' in response:
            status = response["executionStatus"].get('status')
        elif 'eventStatus' in response:
            status = response.get("eventStatus")

        self.state = ExecutionState(status or 'UNKNOWN')
        self.details = response


class RemoteServices:
    """Trigger remote services on a vehicle."""

    def __init__(self, account: "ConnectedDriveAccount", vehicle: "ConnectedDriveVehicle"):
        """Constructor."""
        self._account = account
        self._vehicle = vehicle

    def _get_event_position(self, event_id) -> Dict:
        url = REMOTE_SERVICE_POSITION_URL.format(
            server=self._account.server_url,
            event_id=event_id)
        if self._vehicle.observer_latitude is None or self._vehicle.observer_longitude is None:
            return {
                "errorDetails": {
                    "title": "Unkown position",
                    "description": "Set observer position to retrieve vehicle coordinates!"
                }
            }
        response = self._account.send_request(
            url,
            post=True,
            brand=self._vehicle.brand,
            headers={
                "latitude": str(self._vehicle.observer_latitude),
                "longitude": str(self._vehicle.observer_longitude)
            }
        )
        return response.json()

    def trigger_remote_light_flash(self) -> RemoteServiceStatus:
        """Trigger the vehicle to flash its headlights.

        A state update is NOT triggered after this, as the vehicle state is unchanged.
        """
        _LOGGER.debug('Triggering remote light flash')
        event_id = self._trigger_remote_service(_Services.REMOTE_LIGHT_FLASH)
        return self._block_until_done(_Services.REMOTE_LIGHT_FLASH, event_id)

    def trigger_remote_door_lock(self) -> RemoteServiceStatus:
        """Trigger the vehicle to lock its doors.

        A state update is triggered after this, as the lock state of the vehicle changes.
        """
        _LOGGER.debug('Triggering remote door lock')
        event_id = self._trigger_remote_service(_Services.REMOTE_DOOR_LOCK)
        result = self._block_until_done(_Services.REMOTE_DOOR_LOCK, event_id)
        self._trigger_state_update()
        return result

    def trigger_remote_door_unlock(self) -> RemoteServiceStatus:
        """Trigger the vehicle to unlock its doors.

        A state update is triggered after this, as the lock state of the vehicle changes.
        """
        _LOGGER.debug('Triggering remote door unlock')
        event_id = self._trigger_remote_service(_Services.REMOTE_DOOR_UNLOCK)
        result = self._block_until_done(_Services.REMOTE_DOOR_UNLOCK, event_id)
        self._trigger_state_update()
        return result

    def trigger_remote_horn(self) -> RemoteServiceStatus:
        """Trigger the vehicle to sound its horn.

        A state update is NOT triggered after this, as the vehicle state is unchanged.
        """
        _LOGGER.debug('Triggering remote horn sound')
        event_id = self._trigger_remote_service(_Services.REMOTE_HORN)
        return self._block_until_done(_Services.REMOTE_HORN, event_id)

    def trigger_charge_now(self) -> RemoteServiceStatus:
        """Trigger the vehicle to start charging.

        A state update is NOT triggered after this, as the vehicle state is unchanged.
        """
        _LOGGER.debug('Triggering charge now')
        event_id = self._trigger_remote_service(_Services.REMOTE_CHARGE_NOW)
        result = self._block_until_done(_Services.REMOTE_CHARGE_NOW, event_id)
        self._trigger_state_update()
        return result

    def trigger_remote_air_conditioning(self) -> RemoteServiceStatus:
        """Trigger the air conditioning to start.

        A state update is NOT triggered after this, as the vehicle state is unchanged.
        """
        _LOGGER.debug('Triggering remote air conditioning')
        event_id = self._trigger_remote_service(_Services.REMOTE_AIR_CONDITIONING, action="START")
        result = self._block_until_done(_Services.REMOTE_AIR_CONDITIONING, event_id)
        self._trigger_state_update()
        return result

    def trigger_remote_air_conditioning_stop(self) -> RemoteServiceStatus:
        """Trigger the air conditioning to stop.

        A state update is NOT triggered after this, as the vehicle state is unchanged.
        """
        _LOGGER.debug('Triggering remote air conditioning')
        event_id = self._trigger_remote_service(_Services.REMOTE_AIR_CONDITIONING, action="STOP")
        result = self._block_until_done(_Services.REMOTE_AIR_CONDITIONING, event_id)
        self._trigger_state_update()
        return result

    def _trigger_remote_service(self, service_id: _Services, action=None, post=True) -> str:
        """Trigger a generic remote service.

        You can choose if you want a POST or a GET operation.
        """
        url = REMOTE_SERVICE_URL.format(
            server=self._account.server_url,
            vin=self._vehicle.vin,
            service_type=service_id.value.lower().replace('_', '-')
        )
        params = {"action": action}
        response = self._account.send_request(url, post=post, params=params, brand=self._vehicle.brand)
        try:
            return response.json().get('eventId')
        except JSONDecodeError:
            return None

    def _block_until_done(self, service: _Services = None, event_id: str = None) -> RemoteServiceStatus:
        """Keep polling the server until we get a final answer.

        :raises IOError: if there is no final answer before _POLLING_TIMEOUT
        """
        if not service and not event_id:
            raise ValueError("One of 'service' or 'event_id' is required.")
        fail_after = datetime.datetime.now() + datetime.timedelta(seconds=_POLLING_TIMEOUT)
        while True:
            status = self._get_remote_service_status(service, event_id)
            _LOGGER.debug('current state if remote service is: %s', status.state.value)
            if status.state not in [ExecutionState.UNKNOWN, ExecutionState.PENDING, ExecutionState.DELIVERED]:
                return status
            if datetime.datetime.now() > fail_after:
                raise TimeoutError(
                    'Did not receive remote service result in {} seconds. Current state: {}'.format(
                        _POLLING_TIMEOUT,
                        status.state.value
                    ))
            if status.state == ExecutionState.ERROR:
                raise Exception(
                    "Remote service failed with state '{}'. Response: {}".format(
                        status.state, status.details
                    )
                )
            time.sleep(_POLLING_CYCLE)

    def _get_remote_service_status(self, service: _Services = None, event_id: str = None) -> RemoteServiceStatus:
        """The the execution status of the last remote service that was triggered.

        As the status changes over time, you probably need to poll this.
        Recommended polling time is AT LEAST one second as the reaction is sometimes quite slow.
        """
        if not service and not event_id:
            raise ValueError("One of 'service' or 'event_id' is required.")
        _LOGGER.debug('getting remote service status')
        url = REMOTE_SERVICE_STATUS_URL.format(
            server=self._account.server_url,
            vin=self._vehicle.vin,
            event_id=event_id)
        response = self._account.send_request(url, post=True, brand=self._vehicle.brand)
        try:
            json_result = response.json()
            return RemoteServiceStatus(json_result)
        except ValueError:
            _LOGGER.error('Error decoding json response from the server.')
            _LOGGER.debug(response.headers)
            _LOGGER.debug(response.text)
            raise

    def _trigger_state_update(self) -> None:
        time.sleep(_UPDATE_AFTER_REMOTE_SERVICE_DELAY)
        self._account.update_vehicle_states()

    def trigger_send_poi(self, data: dict) -> RemoteServiceStatus:
        """Send a PointOfInterest to the vehicle.

        :param data: A dictonary containing at least 'lat' and 'lon' and optionally
            'name', 'additionalInfo', 'street', 'city', 'postalCode', 'country',
            'website' or 'phoneNumbers'
        :type data: dict

        A state update is NOT triggered after this, as the vehicle state is unchanged.
        """

        _LOGGER.debug('Sending PointOfInterest to car')
        if "lat" not in data or "lon" not in data:
            raise TypeError("__init__() missing 2 required positional arguments: 'lat' and 'lon'")

        data_json = json.dumps({
            "location": {
                "coordinates": {
                    "latitude": data["lat"],
                    "longitude": data["lon"]
                },
                "name": data.get('name'),
                "locationAddress": {
                    "street": data.get('street'),
                    "postalCode": data.get('postal_code'),
                    "city": data.get('city'),
                    "country": data.get('country'),
                },
                "type": "SHARED_DESTINATION_FROM_EXTERNAL_APP"
            },
            "vin": self._vehicle.vin
        })
        self._account.send_request(
            VEHICLE_POI_URL.format(server=self._account.server_url),
            post=True,
            headers={"Content-Type": "application/json"},
            data=data_json,
            brand=self._vehicle.brand
        )
        # _send_message has no separate ExecutionStates
        return RemoteServiceStatus({'eventStatus': 'EXECUTED'})

    def trigger_remote_vehicle_finder(self) -> RemoteServiceStatus:
        """Trigger the vehicle finder.

        A state update is triggered after this, as the location state of the vehicle changes.
        """
        _LOGGER.debug('Triggering remote vehicle finder')
        # needs to be called via POST, GET is not working
        event_id = self._trigger_remote_service(_Services.REMOTE_VEHICLE_FINDER)
        status = self._block_until_done(_Services.REMOTE_VEHICLE_FINDER, event_id)
        # Sleep another cycle to make sure the results are available
        time.sleep(_POLLING_CYCLE)
        result = self._get_event_position(event_id)
        self._vehicle.status.set_remote_service_position(result)
        return status



"""Get the right url for the different countries."""
from enum import Enum
import logging
from typing import List

_LOGGER = logging.getLogger(__name__)


class Regions(Enum):
    """Regions of the world with separate servers."""
    NORTH_AMERICA = 0
    CHINA = 1
    REST_OF_WORLD = 2


# #: Mapping from regions to servers
# _SERVER_URLS_LEGACY = {
#     Regions.NORTH_AMERICA: "b2vapi.bmwgroup.us",
#     Regions.REST_OF_WORLD: "b2vapi.bmwgroup.com",
#     # Regions.CHINA: "b2vapi.bmwgroup.cn:8592",
# }

_SERVER_URLS_EADRAX = {
    Regions.NORTH_AMERICA: "cocoapi.bmwgroup.us",
    Regions.REST_OF_WORLD: "cocoapi.bmwgroup.com",
    Regions.CHINA: "myprofile.bmw.com.cn",
}

_OCP_APIM_KEYS = {
    Regions.NORTH_AMERICA: "31e102f5-6f7e-7ef3-9044-ddce63891362",
    Regions.REST_OF_WORLD: "4f1c85a3-758f-a37d-bbb6-f8704494acfa",
}

# #: Mapping from regions to servers
# _GCDM_OAUTH_ENDPOINTS = {
#     Regions.NORTH_AMERICA: "login.bmwusa.com/gcdm",
#     Regions.REST_OF_WORLD: "customer.bmwgroup.com/gcdm",
#     # Regions.CHINA: "customer.bmwgroup.cn/gcdm",
# }

# _GCDM_OAUTH_AUTHORIZATION = {
#     Regions.NORTH_AMERICA: {
#         "token": {
#             "Authorization": ("Basic NTQzOTRhNGItYjZjMS00NWZlLWI3YjItOGZkM2FhOTI1M2F"
#                               "hOmQ5MmYzMWMwLWY1NzktNDRmNS1hNzdkLTk2NmY4ZjAwZTM1MQ=="),
#             "code_verifier": "KDarcVUpgymBDCgHDH0PwwMfzycDxu1joeklioOhwXA",
#         },
#         "authenticate": {
#             "client_id": "54394a4b-b6c1-45fe-b7b2-8fd3aa9253aa",
#             "state": "rgastJbZsMtup49-Lp0FMQ",
#         },
#     },
#     Regions.REST_OF_WORLD: {
#         "token": {
#             "Authorization": ("Basic MzFjMzU3YTAtN2ExZC00NTkwLWFhOTktMzNiOTcyNDRkMDQ"
#                               "4OmMwZTMzOTNkLTcwYTItNGY2Zi05ZDNjLTg1MzBhZjY0ZDU1Mg=="),
#             "code_verifier": "7PsmfPS5MpaNt0jEcPpi-B7M7u0gs1Nzw6ex0Y9pa-0",
#         },
#         "authenticate": {
#             "client_id": "31c357a0-7a1d-4590-aa99-33b97244d048",
#             "state": "cEG9eLAIi6Nv-aaCAniziE_B6FPoobva3qr5gukilYw",
#         },
#     },
#     Regions.CHINA: {
#         "token": {
#             "Authorization": ("Basic blF2NkNxdHhKdVhXUDc0eGYzQ0p3VUVQOjF6REh4NnVuNGN"
#                               "EanliTEVOTjNreWZ1bVgya0VZaWdXUGNRcGR2RFJwSUJrN3JPSg=="),
#             "code_verifier": "",
#         },
#         "authenticate": {
#             "client_id": None,
#             "state": None,
#         },
#     },
# }


def valid_regions() -> List[str]:
    """Get list of valid regions as strings."""
    return [region.name.lower() for region in Regions]


def get_region_from_name(name: str) -> Regions:
    """Get a region for a string.

    This function is not case-sensitive.
    """
    for region in Regions:
        if name.lower() == region.name.lower():
            return region
    raise ValueError(
        "Unknown region {}. Valid regions are: {}".format(
            name, ",".join(valid_regions())
        )
    )


# def get_server_url_legacy(region: Regions) -> str:
#     """Get the url of the server for the region."""
#     return _SERVER_URLS_LEGACY[region]


def get_server_url(region: Regions) -> str:
    """Get the url of the server for the region."""
    return _SERVER_URLS_EADRAX[region]


def get_ocp_apim_key(region: Regions) -> str:
    """Get the authorization for OAuth settings."""

    return _OCP_APIM_KEYS[region]

# def get_gcdm_oauth_endpoint(region: Regions) -> str:
#     """Get the url of the server for the region."""
#     return _GCDM_OAUTH_ENDPOINTS[region]


# def get_gcdm_oauth_authorization(region: Regions) -> str:
#     """Get the url of the server for the region."""
#     return _GCDM_OAUTH_AUTHORIZATION[region]
